#
# This file is autogenerated, do not edit manually. Instead,
# edit gen/MotController_data.yml
#

import typing

from hal_impl.data import hal_data

class MotController:

    def __init__(self):
        self.create_called = False

    def create1(self, baseArbId: int) -> None:
        assert isinstance(baseArbId, int)
        self.create_called = True

    def getDeviceNumber(self) -> typing.Tuple[int, int]:
        raise NotImplementedError

    def setDemand(self, mode: int, demand0: int, demand1: int) -> int:
        assert isinstance(mode, int)
        assert isinstance(demand0, int)
        assert isinstance(demand1, int)
        raise NotImplementedError

    def setNeutralMode(self, neutralMode: int) -> None:
        '''Sets the mode of operation during neutral throttle output.

        :param neutralMode:
            The desired mode of operation when the Controller output
            throttle is neutral (ie brake/coast)

        '''
        assert isinstance(neutralMode, int)
        raise NotImplementedError

    def setSensorPhase(self, PhaseSensor: bool) -> None:
        '''Sets the phase of the sensor. Use when controller forward/reverse output
        doesn't correlate to appropriate forward/reverse reading of sensor.

        Pick a value so that positive PercentOutput yields a positive change in sensor.
        After setting this, user can freely call SetInvert() with any value.


        :param PhaseSensor:
            Indicates whether to invert the phase of the sensor.

        '''
        assert isinstance(PhaseSensor, bool)
        raise NotImplementedError

    def setInverted(self, invert: bool) -> None:
        assert isinstance(invert, bool)
        raise NotImplementedError

    def configOpenLoopRamp(self, secondsFromNeutralToFull: float, timeoutMs: int) -> int:
        '''Configures the open-loop ramp rate of throttle output.

        :param secondsFromNeutralToFull:
            Minimum desired time to go from neutral to full throttle. A
            value of '0' will disable the ramp.
        :param timeoutMs:
            Timeout value in ms. Function will generate error if config is
            not successful within timeout.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(secondsFromNeutralToFull, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configClosedLoopRamp(self, secondsFromNeutralToFull: float, timeoutMs: int) -> int:
        '''Configures the closed-loop ramp rate of throttle output.

        :param secondsFromNeutralToFull:
            Minimum desired time to go from neutral to full throttle. A
            value of '0' will disable the ramp.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(secondsFromNeutralToFull, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configPeakOutputForward(self, percentOut: float, timeoutMs: int) -> int:
        '''Configures the forward peak output percentage.

        :param percentOut:
            Desired peak output percentage [0,1].
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(percentOut, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configPeakOutputReverse(self, percentOut: float, timeoutMs: int) -> int:
        '''Configures the reverse peak output percentage.

        :param percentOut:
            Desired peak output percentage.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(percentOut, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configNominalOutputForward(self, percentOut: float, timeoutMs: int) -> int:
        '''Configures the forward nominal output percentage.

        :param percentOut:
            Nominal (minimum) percent output [0,+1].
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(percentOut, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configNominalOutputReverse(self, percentOut: float, timeoutMs: int) -> int:
        '''Configures the reverse nominal output percentage.

        :param percentOut:
            Nominal (minimum) percent output [-1,0].
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(percentOut, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configNeutralDeadband(self, percentDeadband: float, timeoutMs: int) -> int:
        '''Configures the output deadband percentage.

        :param percentDeadband: 
            Desired deadband percentage. Minimum is 0.1%, Maximum is 25%.
            Pass 0.04 for 4% (factory default).
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(percentDeadband, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configVoltageCompSaturation(self, voltage: float, timeoutMs: int) -> int:
        '''Configures the Voltage Compensation saturation voltage.

        :param voltage:
            This is the max voltage to apply to the hbridge when voltage
            compensation is enabled.  For example, if 10 (volts) is specified
            and a TalonSRX is commanded to 0.5 (PercentOutput, closed-loop, etc)
            then the TalonSRX will attempt to apply a duty-cycle to produce 5V.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(voltage, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configVoltageMeasurementFilter(self, filterWindowSamples: int, timeoutMs: int) -> int:
        '''Configures the voltage measurement filter.

        :param filterWindowSamples:
            Number of samples in the rolling average of voltage
            measurement.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(filterWindowSamples, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def enableVoltageCompensation(self, enable: bool) -> None:
        '''Enables voltage compensation. If enabled, voltage compensation works in
        all control modes.

        :param enable:
            Enable state of voltage compensation.

        '''
        assert isinstance(enable, bool)
        raise NotImplementedError

    def getBusVoltage(self) -> typing.Tuple[int, float]:
        '''Gets the bus voltage seen by the device.

        :returns: The bus voltage value (in volts).

        '''
        raise NotImplementedError

    def getMotorOutputPercent(self) -> typing.Tuple[int, float]:
        '''Gets the output percentage of the motor controller.

        :returns: Output of the motor controller (in percent).

        '''
        raise NotImplementedError

    def getOutputCurrent(self) -> typing.Tuple[int, float]:
        '''Gets the output current of the motor controller.

        :returns: The output current (in amps).

        '''
        raise NotImplementedError

    def getTemperature(self) -> typing.Tuple[int, float]:
        '''Gets the temperature of the motor controller.

        :returns: Temperature of the motor controller (in 'C)

        '''
        raise NotImplementedError

    def configSelectedFeedbackSensor(self, feedbackDevice: int, pidIdx: int, timeoutMs: int) -> int:
        '''Select the feedback device for the motor controller.
        Most CTRE CAN motor controllers will support remote sensors over CAN.

        :param feedbackDevice:
            Feedback Device to select.
        :type: feedbackDevice: :class:`.FeedbackDevice` or :class:`.RemoteFeedbackDevice`
        :param pidIdx:
            0 for Primary closed-loop. 1 for cascaded closed-loop.
            See Phoenix-Documentation for how to interpret.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(feedbackDevice, int)
        assert isinstance(pidIdx, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configRemoteFeedbackFilter(self, deviceID: int, remoteSensorSource: int, remoteOrdinal: int, timeoutMs: int) -> int:
        '''Select what remote device and signal to assign to Remote Sensor 0 or Remote Sensor 1.
        After binding a remote device and signal to Remote Sensor X, you may select Remote Sensor X
        as a PID source for closed-loop features.

        :param deviceID:
            The CAN ID of the remote sensor device.
        :param remoteSensorSource:
            The remote sensor device and signal type to bind.
        :param remoteOrdinal:
            0 for configuring Remote Sensor 0
            1 for configuring Remote Sensor 1
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(deviceID, int)
        assert isinstance(remoteSensorSource, int)
        assert isinstance(remoteOrdinal, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configSensorTerm(self, sensorTerm: int, feedbackDevice: int, timeoutMs: int) -> int:
        '''Select what sensor term should be bound to switch feedback device.

        Sensor Sum = Sensor Sum Term 0 - Sensor Sum Term 1

        Sensor Difference = Sensor Diff Term 0 - Sensor Diff Term 1

        The four terms are specified with this routine.  Then Sensor Sum/Difference
        can be selected for closed-looping.

        :param sensorTerm: Which sensor term to bind to a feedback source.
        :param feedbackDevice: The sensor signal to attach to sensorTerm.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(sensorTerm, int)
        assert isinstance(feedbackDevice, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def getSelectedSensorPosition(self, pidIdx: int) -> typing.Tuple[int, int]:
        '''Get the selected sensor position.

        :param pidIdx:
            0 for Primary closed-loop. 1 for cascaded closed-loop.
            See Phoenix-Documentation for how to interpret.
        :returns: Position of selected sensor (in Raw Sensor Units).

        '''
        assert isinstance(pidIdx, int)
        raise NotImplementedError

    def getSelectedSensorVelocity(self, pidIdx: int) -> typing.Tuple[int, int]:
        '''Get the selected sensor velocity.

        :param pidIdx:
            0 for Primary closed-loop. 1 for cascaded closed-loop.
            See Phoenix-Documentation for how to interpret.
        :returns: selected sensor (in raw sensor units) per 100ms.

        '''
        assert isinstance(pidIdx, int)
        raise NotImplementedError

    def setSelectedSensorPosition(self, sensorPos: int, pidIdx: int, timeoutMs: int) -> int:
        '''Sets the sensor position to the given value.

        :param sensorPos:
            Position to set for the selected sensor (in raw sensor units).
        :param pidIdx:
            0 for Primary closed-loop. 1 for cascaded closed-loop.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(sensorPos, int)
        assert isinstance(pidIdx, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def setControlFramePeriod(self, frame: int, periodMs: int) -> int:
        '''Sets the period of the given control frame.

        :param frame:
            Frame whose period is to be changed.
        :param periodMs:
            Period in ms for the given frame.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(frame, int)
        assert isinstance(periodMs, int)
        raise NotImplementedError

    def setStatusFramePeriod(self, frame: int, periodMs: int, timeoutMs: int) -> int:
        '''Sets the period of the given status frame.

        User ensure CAN Bus utilization is not high.
         
        This setting is not persistent and is lost when device is reset. If this
        is a concern, calling application can use HasReset() to determine if the
        status frame needs to be reconfigured.

        :param frame:
            Frame whose period is to be changed.
        :param periodMs:
            Period in ms for the given frame.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(frame, int)
        assert isinstance(periodMs, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def getStatusFramePeriod(self, frame: int, timeoutMs: int) -> typing.Tuple[int, int]:
        '''Gets the period of the given status frame.

        :param frame:
            Frame to get the period of.
        :type frame: :class:`.StatusFrame` or :class:`.StatusFrameEnhanced`
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Period of the given status frame.

        '''
        assert isinstance(frame, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configVelocityMeasurementPeriod(self, period: int, timeoutMs: int) -> int:
        '''Sets the period over which velocity measurements are taken.

        :param period:
            Desired period for the velocity measurement. see :class:`.VelocityMeasPeriod`
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(period, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configVelocityMeasurementWindow(self, windowSize: int, timeoutMs: int) -> int:
        '''Sets the number of velocity samples used in the rolling average velocity
        measurement.

        :param windowSize:
            Number of samples in the rolling average of velocity
            measurement. Valid values are 1,2,4,8,16,32. If another value
            is specified, it will truncate to nearest support value.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(windowSize, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configForwardLimitSwitchSource(self, type: int, normalOpenOrClose: int, deviceID: int, timeoutMs: int) -> int:
        '''Configures the forward limit switch for a remote source. For example, a
        CAN motor controller may need to monitor the Limit-F pin of another Talon
        or CANifier.

        :param type:
            Remote limit switch source. User can choose between a remote
            Talon SRX, CANifier, or deactivate the feature.
        :param normalOpenOrClose:
            Setting for normally open, normally closed, or disabled. This
            setting matches the web-based configuration drop down.
        :param deviceID:
            Device ID of remote source (Talon SRX or CANifier device ID).
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(type, int)
        assert isinstance(normalOpenOrClose, int)
        assert isinstance(deviceID, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configReverseLimitSwitchSource(self, type: int, normalOpenOrClose: int, deviceID: int, timeoutMs: int) -> int:
        '''Configures the reverse limit switch for a remote source. For example, a
        CAN motor controller may need to monitor the Limit-R pin of another Talon
        or CANifier.

        :param type:
            Remote limit switch source. User can choose between a remote
            Talon SRX, CANifier, or deactivate the feature.
        :param normalOpenOrClose:
            Setting for normally open, normally closed, or disabled. This
            setting matches the web-based configuration drop down.
        :param deviceID:
            Device ID of remote source (Talon SRX or CANifier device ID).
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(type, int)
        assert isinstance(normalOpenOrClose, int)
        assert isinstance(deviceID, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def overrideLimitSwitchesEnable(self, enable: bool) -> None:
        '''Sets the enable state for limit switches.

        :param enable:
            Enable state for limit switches.

        '''
        assert isinstance(enable, bool)
        raise NotImplementedError

    def configForwardSoftLimitThreshold(self, forwardSensorLimit: int, timeoutMs: int) -> int:
        '''Configures the forward soft limit threhold.

        :param forwardSensorLimit:
            Forward Sensor Position Limit (in raw Sensor Units).
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(forwardSensorLimit, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configReverseSoftLimitThreshold(self, reverseSensorLimit: int, timeoutMs: int) -> int:
        '''Configures the reverse soft limit threshold.

        :param reverseSensorLimit:
            Reverse Sensor Position Limit (in Raw Sensor Units).
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(reverseSensorLimit, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configForwardSoftLimitEnable(self, enable: bool, timeoutMs: int) -> int:
        '''Configures the forward soft limit enable.

        :param enable:
            Forward Sensor Position Limit Enable.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(enable, bool)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configReverseSoftLimitEnable(self, enable: bool, timeoutMs: int) -> int:
        '''Configures the reverse soft limit enable.

        :param enable:
            Reverse Sensor Position Limit Enable.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(enable, bool)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def overrideSoftLimitsEnable(self, enable: bool) -> None:
        '''Can be used to override-disable the soft limits.
        This function can be used to quickly disable soft limits without
        having to modify the persistent configuration.

        :param enable:
            Enable state for soft limit switches.

        '''
        assert isinstance(enable, bool)
        raise NotImplementedError

    def config_kP(self, slotIdx: int, value: float, timeoutMs: int) -> int:
        '''Sets the 'P' constant in the given parameter slot.

        :param slotIdx:
            Parameter slot for the constant.
        :param value:
            Value of the P constant.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(slotIdx, int)
        assert isinstance(value, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def config_kI(self, slotIdx: int, value: float, timeoutMs: int) -> int:
        '''Sets the 'I' constant in the given parameter slot.

        :param slotIdx:
            Parameter slot for the constant.
        :param value:
            Value of the I constant.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(slotIdx, int)
        assert isinstance(value, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def config_kD(self, slotIdx: int, value: float, timeoutMs: int) -> int:
        '''Sets the 'D' constant in the given parameter slot.

        :param slotIdx:
            Parameter slot for the constant.
        :param value:
            Value of the D constant.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(slotIdx, int)
        assert isinstance(value, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def config_kF(self, slotIdx: int, value: float, timeoutMs: int) -> int:
        '''Sets the 'F' constant in the given parameter slot.

        :param slotIdx:
            Parameter slot for the constant.
        :param value:
            Value of the F constant.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(slotIdx, int)
        assert isinstance(value, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def config_IntegralZone(self, slotIdx: int, izone: float, timeoutMs: int) -> int:
        '''Sets the Integral Zone constant in the given parameter slot. If the
        (absolute) closed-loop error is outside of this zone, integral
        accumulator is automatically cleared. This ensures than integral wind up
        events will stop after the sensor gets far enough from its target.

        :param slotIdx:
            Parameter slot for the constant.
        :param izone:
            Value of the Integral Zone constant. (closed loop error units X 1ms).
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(slotIdx, int)
        assert isinstance(izone, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configAllowableClosedloopError(self, slotIdx: int, allowableClosedLoopError: int, timeoutMs: int) -> int:
        '''Sets the allowable closed-loop error in the given parameter slot.

        :param slotIdx:
            Parameter slot for the constant.
        :param allowableClosedLoopError:
            Value of the allowable closed-loop error.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(slotIdx, int)
        assert isinstance(allowableClosedLoopError, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configMaxIntegralAccumulator(self, slotIdx: int, iaccum: float, timeoutMs: int) -> int:
        '''Sets the maximum integral accumulator in the given parameter slot.

        :param slotIdx:
            Parameter slot for the constant.
        :param iaccum:
            Value to set for the integral accumulator (closed loop error
            units X 1ms).
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(slotIdx, int)
        assert isinstance(iaccum, (int, float))
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def setIntegralAccumulator(self, iaccum: float, pidIdx: int, timeoutMs: int) -> int:
        '''Sets the integral accumulator. Typically this is used to clear/zero the
        integral accumulator, however some use cases may require seeding the
        accumulator for a faster response.

        :param iaccum:
            Value to set for the integral accumulator (closed loop error
            units X 1ms).
        :param pidIdx:
            0 for Primary closed-loop. 1 for cascaded closed-loop.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(iaccum, (int, float))
        assert isinstance(pidIdx, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def getClosedLoopError(self, pidIdx: int) -> typing.Tuple[int, int]:
        '''Gets the closed-loop error.  The units depend on which control mode is in
        use. See Phoenix-Documentation information on units.

        :param slotIdx:
            Parameter slot of the constant.
        :param pidIdx:
            0 for Primary closed-loop. 1 for cascaded closed-loop.
        :returns: Closed-loop error value.

        '''
        assert isinstance(pidIdx, int)
        raise NotImplementedError

    def getIntegralAccumulator(self, pidIdx: int) -> typing.Tuple[int, float]:
        '''Gets the iaccum value.

        :param pidIdx:
            0 for Primary closed-loop. 1 for cascaded closed-loop.
        :returns: Integral accumulator value.

        '''
        assert isinstance(pidIdx, int)
        raise NotImplementedError

    def getErrorDerivative(self, pidIdx: int) -> typing.Tuple[int, float]:
        '''Gets the derivative of the closed-loop error.

        :param slotIdx:
            Parameter slot of the constant.
        :param pidIdx:
            0 for Primary closed-loop. 1 for cascaded closed-loop.
        :returns: The error derivative value.

        '''
        assert isinstance(pidIdx, int)
        raise NotImplementedError

    def selectProfileSlot(self, slotIdx: int, pidIdx: int) -> int:
        '''Selects which profile slot to use for closed-loop control.

        :param slotIdx:
            Profile slot to select.
        :param pidIdx:
            0 for Primary closed-loop. 1 for cascaded closed-loop.

        '''
        assert isinstance(slotIdx, int)
        assert isinstance(pidIdx, int)
        raise NotImplementedError

    def getActiveTrajectoryPosition(self) -> typing.Tuple[int, int]:
        '''Gets the active trajectory target position using
        MotionMagic/MotionProfile control modes.

        :returns: The Active Trajectory Position in sensor units.

        '''
        raise NotImplementedError

    def getActiveTrajectoryVelocity(self) -> typing.Tuple[int, int]:
        '''Gets the active trajectory target velocity using
        MotionMagic/MotionProfile control modes.

        :returns: The Active Trajectory Velocity in sensor units per 100ms.

        '''
        raise NotImplementedError

    def getActiveTrajectoryHeading(self) -> typing.Tuple[int, float]:
        '''Gets the active trajectory target heading using
        MotionMagicArc/MotionProfileArc control modes.

        :returns: The Active Trajectory Heading in degreees.

        '''
        raise NotImplementedError

    def getActiveTrajectoryAll(self) -> typing.Tuple[int, int, int, float]:
        raise NotImplementedError

    def configMotionCruiseVelocity(self, sensorUnitsPer100ms: int, timeoutMs: int) -> int:
        '''Sets the Motion Magic Cruise Velocity.  This is the peak target velocity
        that the motion magic curve generator can use.

        :param sensorUnitsPer100ms:
            Motion Magic Cruise Velocity (in raw Sensor Units per 100 ms).
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(sensorUnitsPer100ms, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configMotionAcceleration(self, sensorUnitsPer100msPerSec: int, timeoutMs: int) -> int:
        '''Sets the Motion Magic Acceleration. This is the target acceleration that
        the motion magic curve generator can use.

        :param sensorUnitsPer100msPerSec:
            Motion Magic Acceleration (in raw sensor units per 100 ms per
            second).
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(sensorUnitsPer100msPerSec, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def clearMotionProfileTrajectories(self) -> int:
        '''Clear the buffered motion profile in both controller's RAM (bottom), and in the
        API (top).

        '''
        raise NotImplementedError

    def getMotionProfileTopLevelBufferCount(self) -> typing.Tuple[int, int]:
        '''Retrieve just the buffer count for the api-level (top) buffer. This
        routine performs no CAN or data structure lookups, so its fast and ideal
        if caller needs to quickly poll the progress of trajectory points being
        emptied into controller's RAM. Otherwise just use GetMotionProfileStatus.

        :returns: number of trajectory points in the top buffer.

        '''
        raise NotImplementedError

    def pushMotionProfileTrajectory(self, position: float, velocity: float, headingDeg: float, profileSlotSelect: int, isLastPoint: bool, zeroPos: bool) -> int:
        '''Push another trajectory point into the top level buffer (which is emptied
        into the motor controller's bottom buffer as room allows).

        :param trajPt: 
            to push into buffer.

            The members should be filled in with these values...
            
            targPos:  servo position in sensor units.
            targVel:  velocity to feed-forward in sensor units
            per 100ms.
            profileSlotSelect  which slot to pull PIDF gains from.  Currently
            supports 0,1,2,3.
            isLastPoint  set to nonzero to signal motor controller to keep processing this
            trajectory point, instead of jumping to the next one
            when timeDurMs expires.  Otherwise MP executer will
            eventually see an empty buffer after the last point
            expires, causing it to assert the IsUnderRun flag.
            However this may be desired if calling application
            never wants to terminate the MP.
            zeroPos  set to nonzero to signal motor controller to "zero" the selected
            position sensor before executing this trajectory point.
            Typically the first point should have this set only thus
            allowing the remainder of the MP positions to be relative to
            zero.
        :returns: 
            CTR_OKAY if trajectory point push ok. ErrorCode if buffer is
            full due to kMotionProfileTopBufferCapacity.

        '''
        assert isinstance(position, (int, float))
        assert isinstance(velocity, (int, float))
        assert isinstance(headingDeg, (int, float))
        assert isinstance(profileSlotSelect, int)
        assert isinstance(isLastPoint, bool)
        assert isinstance(zeroPos, bool)
        raise NotImplementedError

    def isMotionProfileTopLevelBufferFull(self) -> typing.Tuple[int, bool]:
        '''Retrieve just the buffer full for the api-level (top) buffer. This
        routine performs no CAN or data structure lookups, so its fast and ideal
        if caller needs to quickly poll. Otherwise just use
        GetMotionProfileStatus.

        :returns: number of trajectory points in the top buffer.

        '''
        raise NotImplementedError

    def processMotionProfileBuffer(self) -> int:
        '''This must be called periodically to funnel the trajectory points from the
        API's top level buffer to the controller's bottom level buffer. Recommendation
        is to call this twice as fast as the execution rate of the motion
        profile. So if MP is running with 20ms trajectory points, try calling
        this routine every 10ms. All motion profile functions are thread-safe
        through the use of a mutex, so there is no harm in having the caller
        utilize threading.

        '''
        raise NotImplementedError

    def getMotionProfileStatus(self) -> typing.Tuple[int, int, int, int, bool, bool, bool, bool, int, int]:
        '''Retrieve all status information.
        For best performance, Caller can snapshot all status information regarding the
        motion profile executer.

        :param statusToFill: 
            Caller supplied object to fill.

            The members are filled, as follows...
            
            topBufferRem:   The available empty slots in the trajectory buffer.
            The robot API holds a "top buffer" of trajectory points, so your applicaion
            can dump several points at once.  The API will then stream them into the
            low-level buffer, allowing the motor controller to act on them.
            
            topBufferRem: The number of points in the top trajectory buffer.
            
            btmBufferCnt: The number of points in the low level controller buffer.
            
            hasUnderrun:    Set if isUnderrun ever gets set.
            Only is cleared by clearMotionProfileHasUnderrun() to ensure
            
            isUnderrun:     This is set if controller needs to shift a point from its buffer into
            the active trajectory point however
            the buffer is empty.
            This gets cleared automatically when is resolved.
            
            activePointValid:   True if the active trajectory point has not empty, false otherwise. The members in activePoint are only valid if this signal is set.
            
            isLast: is set/cleared based on the MP executer's current
            trajectory point's IsLast value.  This assumes
            IsLast was set when PushMotionProfileTrajectory
            was used to insert the currently processed trajectory
            point.
            
            profileSlotSelect: The currently processed trajectory point's
            selected slot.  This can differ in the currently selected slot used
            for Position and Velocity servo modes
            
            outputEnable:       The current output mode of the motion profile
            executer (disabled, enabled, or hold).  When changing the set()
            value in MP mode, it's important to check this signal to
            confirm the change takes effect before interacting with the top buffer.

        '''
        raise NotImplementedError

    def clearMotionProfileHasUnderrun(self, timeoutMs: int) -> int:
        '''Clear the "Has Underrun" flag. Typically this is called after application
        has confirmed an underrun had occured.

        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def changeMotionControlFramePeriod(self, periodMs: int) -> int:
        '''Calling application can opt to speed up the handshaking between the robot
        API and the controller to increase the download rate of the controller's Motion
        Profile. Ideally the period should be no more than half the period of a
        trajectory point.

        :param periodMs:
            The transmit period in ms.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(periodMs, int)
        raise NotImplementedError

    def getLastError(self) -> int:
        '''Gets the last error generated by this object.

        :returns: Last Error Code generated by a function.

        '''
        raise NotImplementedError

    def getFirmwareVersion(self) -> typing.Tuple[int, int]:
        '''Gets the firmware version of the device.

        :returns: Firmware version of device. For example: version 1-dot-2 
            is 0x0102.

        '''
        raise NotImplementedError

    def hasResetOccurred(self) -> typing.Tuple[int, bool]:
        ''':returns: Returns true if the device has reset since last call

        '''
        raise NotImplementedError

    def configSetCustomParam(self, newValue: int, paramIndex: int, timeoutMs: int) -> int:
        '''Sets the value of a custom parameter. This is for arbitrary use.

        Sometimes it is necessary to save calibration/limit/target information 
        in the device. Particularly if the device is part of a subsystem that 
        can be replaced.

        :param newValue:
            Value for custom parameter.
        :param paramIndex:
            Index of custom parameter [0,1].
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(newValue, int)
        assert isinstance(paramIndex, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configGetCustomParam(self, paramIndex: int, timoutMs: int) -> typing.Tuple[int, int]:
        '''Gets the value of a custom parameter.

        :param paramIndex:
            Index of custom parameter [0,1].
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Value of the custom param.

        '''
        assert isinstance(paramIndex, int)
        assert isinstance(timoutMs, int)
        raise NotImplementedError

    def configSetParameter(self, param: int, value: float, subValue: int, ordinal: int, timeoutMs: int) -> int:
        '''Sets a parameter.

        :param param:
            Parameter enumeration.
        :param value:
            Value of parameter.
        :param subValue:
            Subvalue for parameter. Maximum value of 255.
        :param ordinal:
            Ordinal of parameter.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(param, int)
        assert isinstance(value, (int, float))
        assert isinstance(subValue, int)
        assert isinstance(ordinal, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configGetParameter(self, param: int, ordinal: int, timeoutMs: int) -> typing.Tuple[int, float]:
        '''Gets a parameter.

        :param param:
            Parameter enumeration.
        :param ordinal:
            Ordinal of parameter.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.

        :returns: Value of parameter.

        '''
        assert isinstance(param, int)
        assert isinstance(ordinal, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configPeakCurrentLimit(self, amps: int, timeoutMs: int) -> int:
        '''Configure the peak allowable current (when current limit is enabled).

        Current limit is activated when current exceeds the peak limit for longer
        than the peak duration. Then software will limit to the continuous limit.
        This ensures current limiting while allowing for momentary excess current
        events.

        For simpler current-limiting (single threshold) use
        ConfigContinuousCurrentLimit() and set the peak to zero:
        ConfigPeakCurrentLimit(0).

        Not available on the Victor SPX

        :param amps:
            Amperes to limit.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.

        '''
        assert isinstance(amps, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configPeakCurrentDuration(self, milliseconds: int, timeoutMs: int) -> int:
        '''Configure the peak allowable duration (when current limit is enabled).

        Current limit is activated when current exceeds the peak limit for longer
        than the peak duration. Then software will limit to the continuous limit.
        This ensures current limiting while allowing for momentary excess current
        events.

        For simpler current-limiting (single threshold) use
        ConfigContinuousCurrentLimit() and set the peak to zero:
        ConfigPeakCurrentLimit(0).

        Not available on the Victor SPX

        :param milliseconds:
            How long to allow current-draw past peak limit.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.

        '''
        assert isinstance(milliseconds, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configContinuousCurrentLimit(self, amps: int, timeoutMs: int) -> int:
        '''Configure the continuous allowable current-draw (when current limit is
        enabled).

        Current limit is activated when current exceeds the peak limit for longer
        than the peak duration. Then software will limit to the continuous limit.
        This ensures current limiting while allowing for momentary excess current
        events.

        For simpler current-limiting (single threshold) use
        ConfigContinuousCurrentLimit() and set the peak to zero:
        ConfigPeakCurrentLimit(0).

        Not available on the Victor SPX

        :param amps:
            Amperes to limit.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.

        '''
        assert isinstance(amps, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def enableCurrentLimit(self, enable: bool) -> int:
        '''Enable or disable Current Limit.

        :param enable:
            Enable state of current limit.

        Not available on the Victor SPX

        see :meth:`.configPeakCurrentLimit`, :meth:`.configPeakCurrentDuration`, 
        :meth:`.configContinuousCurrentLimit`

        '''
        assert isinstance(enable, bool)
        raise NotImplementedError

    def setLastError(self, error: int) -> int:
        assert isinstance(error, int)
        raise NotImplementedError

    def getAnalogIn(self) -> typing.Tuple[int, int]:
        '''Get the position of whatever is in the analog pin of the Talon,
        regardless of whether it is actually being used for feedback.

        :returns: 
            the 24bit analog value. The bottom ten bits is the ADC (0 - 1023)
            on the analog pin of the Talon. The upper 14 bits tracks the
            overflows and underflows (continuous sensor).

        '''
        raise NotImplementedError

    def setAnalogPosition(self, newPosition: int, timeoutMs: int) -> int:
        '''Sets analog position.

        :param newPosition:
            The new position.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: an ErrorCode.

        '''
        assert isinstance(newPosition, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def getAnalogInRaw(self) -> typing.Tuple[int, int]:
        '''Get the position of whatever is in the analog pin of the Talon,
        regardless of whether it is actually being used for feedback.

        :returns: the ADC (0 - 1023) on analog pin of the Talon.

        '''
        raise NotImplementedError

    def getAnalogInVel(self) -> typing.Tuple[int, int]:
        '''Get the velocity of whatever is in the analog pin of the Talon,
        regardless of whether it is actually being used for feedback.

        :returns: the speed in units per 100ms where 1024 units is one rotation.

        '''
        raise NotImplementedError

    def getQuadraturePosition(self) -> typing.Tuple[int, int]:
        '''Get the quadrature position of the Talon, regardless of whether
        it is actually being used for feedback.

        :returns: the Error code of the request.

        '''
        raise NotImplementedError

    def setQuadraturePosition(self, newPosition: int, timeoutMs: int) -> int:
        '''Change the quadrature reported position. Typically this is used to "zero"
        the sensor. This only works with Quadrature sensor. To set the selected
        sensor position regardless of what type it is, see
        SetSelectedSensorPosition in the motor controller class.

        :param newPosition:
            The position value to apply to the sensor.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: error code.

        '''
        assert isinstance(newPosition, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def getQuadratureVelocity(self) -> typing.Tuple[int, int]:
        '''Get the quadrature velocity, regardless of whether
        it is actually being used for feedback.

        :returns: the value (0 - 1023) on the analog pin of the Talon.

        '''
        raise NotImplementedError

    def getPulseWidthPosition(self) -> typing.Tuple[int, int]:
        '''Gets pulse width position, regardless of whether
        it is actually being used for feedback.

        :returns: the pulse width position.

        '''
        raise NotImplementedError

    def setPulseWidthPosition(self, newPosition: int, timeoutMs: int) -> int:
        '''Sets pulse width position.

        :param newPosition:
            The position value to apply to the sensor.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: an ErrErrorCode

        '''
        assert isinstance(newPosition, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def getPulseWidthVelocity(self) -> typing.Tuple[int, int]:
        '''Gets pulse width velocity, regardless of whether
        it is actually being used for feedback.

        :returns: the pulse width velocity  in units per 100ms (where 4096 units is 1 rotation).

        '''
        raise NotImplementedError

    def getPulseWidthRiseToFallUs(self) -> typing.Tuple[int, int]:
        '''Gets pulse width rise to fall time.

        :returns: the pulse width rise to fall time in microseconds.

        '''
        raise NotImplementedError

    def getPulseWidthRiseToRiseUs(self) -> typing.Tuple[int, int]:
        '''Gets pulse width rise to rise time.

        :returns: the pulse width rise to rise time in microseconds.

        '''
        raise NotImplementedError

    def getPinStateQuadA(self) -> typing.Tuple[int, int]:
        '''
        '''
        raise NotImplementedError

    def getPinStateQuadB(self) -> typing.Tuple[int, int]:
        raise NotImplementedError

    def getPinStateQuadIdx(self) -> typing.Tuple[int, int]:
        raise NotImplementedError

    def isFwdLimitSwitchClosed(self) -> typing.Tuple[int, int]:
        raise NotImplementedError

    def isRevLimitSwitchClosed(self) -> typing.Tuple[int, int]:
        raise NotImplementedError

    def getFaults(self) -> typing.Tuple[int, int]:
        raise NotImplementedError

    def getStickyFaults(self) -> typing.Tuple[int, int]:
        raise NotImplementedError

    def clearStickyFaults(self, timeoutMs: int) -> int:
        '''Clears all sticky faults.

        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Last Error Code generated by a function.

        '''
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def selectDemandType(self, enable: bool) -> int:
        '''For now this simply updates the CAN signal to the motor controller.
        Future firmware updates will use this to control advanced cascaded loop behavior.

        :param value:

        '''
        assert isinstance(enable, bool)
        raise NotImplementedError

    def setMPEOutput(self, MpeOutput: int) -> int:
        assert isinstance(MpeOutput, int)
        raise NotImplementedError

    def enableHeadingHold(self, enable: bool) -> int:
        '''Enables a future feature called "Heading Hold".
        For now this simply updates the CAN signal to the motor controller.
        Future firmware updates will use this.

        :param enable: true/false enable

        '''
        assert isinstance(enable, bool)
        raise NotImplementedError

    def getAnalogInAll(self) -> typing.Tuple[int, int, int, int]:
        raise NotImplementedError

    def getQuadratureSensor(self) -> typing.Tuple[int, int, int]:
        raise NotImplementedError

    def getPulseWidthAll(self) -> typing.Tuple[int, int, int, int, int]:
        raise NotImplementedError

    def getQuadPinStates(self) -> typing.Tuple[int, int, int, int]:
        raise NotImplementedError

    def getLimitSwitchState(self) -> typing.Tuple[int, int, int]:
        raise NotImplementedError





from .wpi_talonsrx import WPI_TalonSRX
from .wpi_victorspx import WPI_VictorSPX
from .canifier import CANifier
from .pigeonimu import PigeonIMU

try:
    from .version import __version__
except ImportError:     # pragma: nocover
    __version__ = 'master'



# validated: 2018-01-06 EN d090f9cdc6a1 libraries/driver/include/ctre/phoenix/Motion/TrajectoryPoint.h
from collections import namedtuple


__all__ = ['TrajectoryPoint']


#: Motion Profile Trajectory Point
#: This is simply a data transfer object.
#:
#: :param position: The position to servo to.
#: :param velocity: The velocity to feed-forward.
#: :param headingDeg:
#: :param profileSlotSelect:
#:      Which slot to get PIDF gains.
#:      PID is used for position servo.
#:      F is used as the Kv constant for velocity feed-forward.
#:      Typically this is hardcoded to a particular slot, but you are free to
#:      gain schedule if need be.
#: :param isLastPoint:
#:      Set to true to signal Talon that this is the final point, so do not
#:      attempt to pop another trajectory point from out of the Talon buffer.
#:      Instead continue processing this way point.  Typically the velocity
#:      member variable should be zero so that the motor doesn't spin indefinitely.
#: :param zeroPos:
#:      Set to true to signal Talon to zero the selected sensor.
#:      When generating MPs, one simple method is to make the first target position zero,
#:      and the final target position the target distance from the current position.
#:      Then when you fire the MP, the current position gets set to zero.
#:      If this is the intent, you can set zeroPos on the first trajectory point.
#:
#:      Otherwise you can leave this false for all points, and offset the positions
#:      of all trajectory points so they are correct.
TrajectoryPoint = namedtuple("TrajectoryPoint", ["position", "velocity", "headingDeg", "profileSlotSelect",
                                                 "isLastPoint", "zeroPos"])



# validated: 2018-01-06 EN d090f9cdc6a1 libraries/driver/include/ctre/phoenix/Motion/MotionProfileStatus.h
from collections import namedtuple


__all__ = ['MotionProfileStatus']


#: Motion Profile Status
#: This is simply a data transfer object.
#:
#: :param topBufferRem:
#:      The available empty slots in the trajectory buffer.
#:     
#:      The robot API holds a "top buffer" of trajectory points, so your applicaion
#:      can dump several points at once.  The API will then stream them into the Talon's
#:      low-level buffer, allowing the Talon to act on them.
#: :param topBufferCnt:
#:      The number of points in the top trajectory buffer.
#: :param btmBufferCnt:
#:      The number of points in the low level Talon buffer.
#: :param hasUnderrun:
#:      Set if isUnderrun ever gets set.
#:      Only is cleared by clearMotionProfileHasUnderrun() to ensure
#:      robot logic can react or instrument it.
#:      @see clearMotionProfileHasUnderrun()
#: :param isUnderrun:
#:      This is set if Talon needs to shift a point from its buffer into
#:      the active trajectory point however the buffer is empty. This gets cleared
#:      automatically when is resolved.
#: :param activePointValid:
#:      True if the active trajectory point has not empty, false otherwise.
#:      The members in activePoint are only valid if this signal is set.
#: :param isLast:
#: :param profileSlotSelect:
#: :param outputEnable:
#:      The current output mode of the motion profile executer (disabled, enabled, or hold).
#:      When changing the set() value in MP mode, it's important to check this signal to
#:      confirm the change takes effect before interacting with the top buffer.
MotionProfileStatus = namedtuple("MotionProfileStatus", ["topBufferRem", "topBufferCnt", "btmBufferCnt", "hasUnderrun",
                                                         "isUnderrun", "activePointValid", "isLast", "profileSlotSelect"])



# validated: 2018-01-05 EN 8f8595a20145 java/src/com/ctre/phoenix/sensors/PigeonIMU.java
#----------------------------------------------------------------------------
#  Software License Agreement
#
# Copyright (C) Cross The Road Electronics.  All rights
# reserved.
# 
# Cross The Road Electronics (CTRE) licenses to you the right to 
# use, publish, and distribute copies of CRF (Cross The Road) firmware files (*.crf) and Software
# API Libraries ONLY when in use with Cross The Road Electronics hardware products.
# 
# THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT
# LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL
# CROSS THE ROAD ELECTRONICS BE LIABLE FOR ANY INCIDENTAL, SPECIAL, 
# INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF
# PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
# BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE
# THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER
# SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT
# (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE
#----------------------------------------------------------------------------
import enum
import typing

import hal
from collections import namedtuple
from wpilib._impl.utils import match_arglist
from .talonsrx import TalonSRX
from ._impl import (
    ErrorCode,
    ParamEnum,
    PigeonIMU_StatusFrame,
    PigeonIMU as PigeonImuImpl,
    PigeonIMU_ControlFrame,
    PigeonIMU_Faults,
    PigeonIMU_StickyFaults,
)


__all__ = ['PigeonIMU', 'FusionStatus', 'GeneralStatus', 'PigeonState', 'CalibrationMode']


_FusionStatus = namedtuple("_FusionStatus", ["bIsFusing", "bIsValid", "heading", "lastError"])

class FusionStatus(_FusionStatus):
    """Data object for holding fusion information."""
    def __str__(self):
        if self.lastError != ErrorCode.OK:
            description = "Could not receive status frame.  Check wiring and web-config."
        elif not self.bIsValid:
            description = "Fused Heading is not valid."
        elif not self.bIsFusing:
            description = "Fused Heading is valid."
        else:
            description = "Fused Heading is valid and is fusing compass."
        return description


class CalibrationMode(enum.IntEnum):
    """Various calibration modes supported by Pigeon."""
    BootTareGyroAccel = 0
    Temperature = 1
    Magnetometer12Pt = 2
    Magnetometer360 = 3
    Accelerometer = 5
    Unknown = -1


class PigeonState(enum.IntEnum):
    """Overall state of the Pigeon."""
    NoComm = 0
    Initializing = 1
    Ready = 2
    UserCalibration = 3
    Unknown = -1


_GeneralStatus = namedtuple('_GeneralStatus', [
    'state', 'currentMode', 'calibrationError', 'bCalIsBooting', 'tempC', 'upTimeSec', 
    'noMotionBiasCount', 'tempCompensationCount', 'lastError'])


class GeneralStatus(_GeneralStatus):
    """Data object for status on current calibration and general status.
    
    Pigeon has many calibration modes supported for a variety of uses. The
    modes generally collects and saves persistently information that makes
    the Pigeon signals more accurate. This includes collecting temperature,
    gyro, accelerometer, and compass information.
    
    For FRC use-cases, typically compass and temperature calibration is not
    required.
    
    Additionally when motion driver software in the Pigeon boots, it will
    perform a fast boot calibration to initially bias gyro and setup
    accelerometer.
    
    These modes can be enabled with the EnterCalibration mode.
    
    When a calibration mode is entered, caller can expect...
    
    - 
        PigeonState to reset to Initializing and bCalIsBooting is set to true.
        Pigeon LEDs will blink the boot pattern. This is similar to the normal
        boot cal, however it can an additional ~30 seconds since calibration
        generally requires more information. currentMode will reflect the user's
        selected calibration mode.
    
    - 
        PigeonState will eventually settle to UserCalibration and Pigeon LEDs
        will show cal specific blink patterns. bCalIsBooting is now false.
    
    - 
        Follow the instructions in the Pigeon User Manual to meet the
        calibration specific requirements. When finished calibrationError will
        update with the result. Pigeon will solid-fill LEDs with red (for
        failure) or green (for success) for ~5 seconds. Pigeon then perform
        boot-cal to cleanly apply the newly saved calibration data.

    :param state:
        The current state of the motion driver. This reflects if the sensor
        signals are accurate. Most calibration modes will force Pigeon to
        reinit the motion driver.
    :param currentMode:
        The currently applied calibration mode if state is in UserCalibration
        or if bCalIsBooting is true. Otherwise it holds the last selected
        calibration mode (when calibrationError was updated).
    :param calibrationError:
        The error code for the last calibration mode. Zero represents a
        successful cal (with solid green LEDs at end of cal) and nonzero is a
        failed calibration (with solid red LEDs at end of cal). Different
        calibration
    :param bCalIsBooting:
        After caller requests a calibration mode, pigeon will perform a
        boot-cal before entering the requested mode. During this period, this
        flag is set to true.
    :param tempC:
        Temperature in Celsius
    :param upTimeSec:
        Number of seconds Pigeon has been up (since boot). This register is
        reset on power boot or processor reset. Register is capped at 255
        seconds with no wrap around.
    :param noMotionBiasCount:
        Number of times the Pigeon has automatically rebiased the gyro. This
        counter overflows from 15 -> 0 with no cap.
    :param tempCompensationCount:
        Number of times the Pigeon has temperature compensated the various
        signals. This counter overflows from 15 -> 0 with no cap.
    :param lastError:
        Same as getLastError()
    """
    def __str__(self):
        """
        general string description of current status"""
        if self.lastError != ErrorCode.OK: # same as NoComm
            description = "Status frame was not received, check wired connections and web-based config."
        elif self.bCalIsBooting:
            description = "Pigeon is boot-caling to properly bias accel and gyro.  Do not move Pigeon.  When finished biasing, calibration mode will start."
        elif self.state == PigeonState.UserCalibration:
            # mode specific descriptions 
            if self.currentMode == CalibrationMode.BootTareGyroAccel:
                description = "Boot-Calibration: Gyro and Accelerometer are being biased."
            elif self.currentMode == CalibrationMode.Temperature:
                description = "Temperature-Calibration: Pigeon is collecting temp data and will finish when temp range is reached. \n"
                description += "Do not move Pigeon."
            elif self.currentMode == CalibrationMode.Magnetometer12Pt:
                description = "Magnetometer Level 1 calibration: Orient the Pigeon PCB in the 12 positions documented in the User's Manual."
            elif self.currentMode == CalibrationMode.Magnetometer360:
                description = "Magnetometer Level 2 calibration: Spin robot slowly in 360' fashion."
            elif self.currentMode == CalibrationMode.Accelerometer:
                description = "Accelerometer Calibration: Pigeon PCB must be placed on a level source.  Follow User's Guide for how to level surface."
            else:
                description = "Unknown status"
        elif self.state == PigeonState.Ready:
            # definitely not doing anything cal-related. So just instrument
            # the motion driver state
            description = "Pigeon is running normally.  Last CAL error code was %s." % (self.calibrationError,)
        elif self.state == PigeonState.Initializing:
            # definitely not doing anything cal-related. So just instrument
            # the motion driver state
            description = "Pigeon is boot-caling to properly bias accel and gyro.  Do not move Pigeon."
        else:
            description = "Not enough data to determine status."

        return description


class PigeonIMU(PigeonImuImpl):
    """Pigeon IMU Class. Class supports communicating over CANbus and over
    ribbon-cable (CAN Talon SRX).
    """

    def __init__(self, *args, **kwargs):
        """
        Arguments can be structured as follows:

        - deviceNumber
        - talonSrx
        
        :param deviceNumber:
            CAN Device Id of Pigeon [0,62]
        :param talonSrx:
            Object for the TalonSRX connected via ribbon cable.
        """
        super().__init__()

        deviceNumber_arg = ("deviceNumber", [int])
        talonSrx_arg = ("talonSrx", [TalonSRX])
        templates = [[deviceNumber_arg], [talonSrx_arg]]
        index, results = match_arglist('PIDController.__init__',
                                   args, kwargs, templates)
        self.generalStatus = [0] * 10
        self.fusionStatus = [0] * 10

        if index == 0:
            self.deviceNumber = results['deviceNumber']
            self.create1(self.deviceNumber)
        elif index == 1:
            self.deviceNumber = results['talonSrx'].getDeviceID()
            self.create2(self.deviceNumber)
            hal.report(64, self.deviceNumber + 1)
        hal.report(hal.UsageReporting.kResourceType_PigeonIMU, self.deviceNumber + 1)

    def getGeneralStatus(self) -> GeneralStatus:
        """
        Get the status of the current (or previousley complete) calibration.
        
        :returns: :class:`.GeneralStatus`
            generalstatus.lastError is Error Code generated by function.  0 indicates no error.
        """
        results = super().getGeneralStatus()
        return GeneralStatus(results)

    def getFusedHeading(self) -> float:
        """
        :param status:
            object reference to fill with fusion status flags.
            Caller may pass null if flags are not needed.
        :returns: :class:`.FusionStatus`
        """
        results = self.getFusedHeading2()
        return FusionStatus(results)
        
    def getFaults(self) -> typing.Tuple[int, PigeonIMU_Faults]:
        """
        Gets the fault status

        :returns: (error code, :class:`.PigeonIMU_Faults`)
            Error Code generated by function. 0 indicates no error.
        """
        _, bits = super().getFaults()
        return self.getLastError(), PigeonIMU_Faults(bits)

    def getStickyFaults(self) -> typing.Tuple[int, PigeonIMU_StickyFaults]:
        """
        Gets the sticky fault status

        :returns: (error code, :class:`.PigeonIMU_StickyFaults)
            Error Code generated by function. 0 indicates no error.
        """
        _, bits = super().getStickyFaults()
        return self.getLastError(), PigeonIMU_StickyFaults(bits)



#
# This file is autogenerated, do not edit manually. Instead,
# edit gen/CANifier_data.yml
#

import typing

from hal_impl.data import hal_data

class CANifier:

    def __init__(self):
        self.create_called = False

    def create1(self, deviceNumber: int) -> None:
        assert isinstance(deviceNumber, int)
        self.create_called = True

    def setLEDOutput(self, dutyCycle: int, ledChannel: int) -> int:
        '''Sets the LED Output

        :param percentOutput: Output duty cycle expressed as percentage.
        :param ledChannel: Channel to set the output of.

        '''
        assert isinstance(dutyCycle, int) and dutyCycle < 4294967296 and dutyCycle >= 0
        assert isinstance(ledChannel, int) and ledChannel < 4294967296 and ledChannel >= 0
        raise NotImplementedError

    def setGeneralOutputs(self, outputsBits: int, isOutputBits: int) -> int:
        '''Sets the output of all General Pins

        :param outputBits: A bit mask of all the output states.  LSB->MSB is in the order of the #GeneralPin enum.
        :param isOutputBits: A boolean bit mask that sets the pins to be outputs or inputs.  A bit of 1 enables output.

        '''
        assert isinstance(outputsBits, int) and outputsBits < 4294967296 and outputsBits >= 0
        assert isinstance(isOutputBits, int) and isOutputBits < 4294967296 and isOutputBits >= 0
        raise NotImplementedError

    def setGeneralOutput(self, outputPin: int, outputValue: bool, outputEnable: bool) -> int:
        '''Sets the output of a General Pin

        :param outputPin: The pin to use as output.
        :param outputValue: The desired output state.
        :param outputEnable: Whether this pin is an output. "True" enables output.

        '''
        assert isinstance(outputPin, int) and outputPin < 4294967296 and outputPin >= 0
        assert isinstance(outputValue, bool)
        assert isinstance(outputEnable, bool)
        raise NotImplementedError

    def setPWMOutput(self, pwmChannel: int, dutyCycle: int) -> int:
        '''
        '''
        assert isinstance(pwmChannel, int) and pwmChannel < 4294967296 and pwmChannel >= 0
        assert isinstance(dutyCycle, int) and dutyCycle < 4294967296 and dutyCycle >= 0
        raise NotImplementedError

    def enablePWMOutput(self, pwmChannel: int, bEnable: bool) -> int:
        assert isinstance(pwmChannel, int) and pwmChannel < 4294967296 and pwmChannel >= 0
        assert isinstance(bEnable, bool)
        raise NotImplementedError

    def getGeneralInput(self, inputPin: int) -> typing.Tuple[int, bool]:
        '''Gets the state of the specified pin

        :param inputPin: The index of the pin.
        :returns: The state of the pin.

        '''
        assert isinstance(inputPin, int) and inputPin < 4294967296 and inputPin >= 0
        raise NotImplementedError

    def getPWMInput(self, pwmChannel: int) -> typing.Tuple[int, list]:
        '''Gets the PWM Input

        :param pwmChannel: PWM channel to get.
        :param dutyCycleAndPeriod: Double array to hold Duty Cycle [0] and Period [1].

        '''
        assert isinstance(pwmChannel, int) and pwmChannel < 4294967296 and pwmChannel >= 0
        raise NotImplementedError

    def getLastError(self) -> int:
        '''Call GetLastError() generated by this object.
        Not all functions return an error code but can
        potentially report errors.

        This function can be used to retrieve those error codes.

        :returns: The last ErrorCode generated.

        '''
        raise NotImplementedError

    def getBusVoltage(self) -> typing.Tuple[int, float]:
        '''Gets the bus voltage seen by the device.

        :returns: The bus voltage value (in volts).

        '''
        raise NotImplementedError

    def setLastError(self, error: int) -> None:
        assert isinstance(error, int)
        raise NotImplementedError

    def configSetParameter(self, param: int, value: float, subValue: int, ordinal: int, timeoutMs: int) -> int:
        '''Sets a parameter. Generally this is not used.

        This can be utilized in

        - Using new features without updating API installation.
        - Errata workarounds to circumvent API implementation.
        - Allows for rapid testing / unit testing of firmware.

        :param param:
            Parameter enumeration.
        :param value:
            Value of parameter.
        :param subValue:
            Subvalue for parameter. Maximum value of 255.
        :param ordinal:
            Ordinal of parameter.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(param, int)
        assert isinstance(value, (int, float))
        assert isinstance(subValue, int)
        assert isinstance(ordinal, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configGetParameter(self, param: int, ordinal: int, timeoutMs: int) -> typing.Tuple[int, float]:
        '''Gets a parameter. Generally this is not used.

        This can be utilized in

        - Using new features without updating API installation.
        - Errata workarounds to circumvent API implementation.
        - Allows for rapid testing / unit testing of firmware.

        :param param:
            Parameter enumeration.
        :param ordinal:
            Ordinal of parameter.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Value of parameter.

        '''
        assert isinstance(param, int)
        assert isinstance(ordinal, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configSetCustomParam(self, newValue: int, paramIndex: int, timeoutMs: int) -> int:
        '''Sets the value of a custom parameter. This is for arbitrary use.

        Sometimes it is necessary to save calibration/duty cycle/output
        information in the device. Particularly if the
        device is part of a subsystem that can be replaced.

        :param newValue:
            Value for custom parameter.
        :param paramIndex:
            Index of custom parameter [0-1].
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(newValue, int)
        assert isinstance(paramIndex, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def configGetCustomParam(self, paramIndex: int, timoutMs: int) -> typing.Tuple[int, int]:
        '''Gets the value of a custom parameter. This is for arbitrary use.

        Sometimes it is necessary to save calibration/duty cycle/output
        information in the device. Particularly if the
        device is part of a subsystem that can be replaced.

        :param paramIndex:
            Index of custom parameter [0-1].
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Value of the custom param.

        '''
        assert isinstance(paramIndex, int)
        assert isinstance(timoutMs, int)
        raise NotImplementedError

    def getFaults(self) -> typing.Tuple[int, int]:
        raise NotImplementedError

    def getStickyFaults(self) -> typing.Tuple[int, int]:
        raise NotImplementedError

    def clearStickyFaults(self, timeoutMs: int) -> int:
        '''Clears the Sticky Faults

        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def getFirmwareVersion(self) -> typing.Tuple[int, int]:
        '''Gets the firmware version of the device.

        :returns: Firmware version of device.

        '''
        raise NotImplementedError

    def hasResetOccurred(self) -> typing.Tuple[int, bool]:
        '''Returns true if the device has reset since last call.

        :returns: Has a Device Reset Occurred?

        '''
        raise NotImplementedError

    def setStatusFramePeriod(self, frame: int, periodMs: int, timeoutMs: int) -> int:
        '''Sets the period of the given status frame.

        :param statusFrame:
            Frame whose period is to be changed.
        :param periodMs:
            Period in ms for the given frame.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(frame, int)
        assert isinstance(periodMs, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def getStatusFramePeriod(self, frame: int, timeoutMs: int) -> typing.Tuple[int, int]:
        '''Gets the period of the given status frame.

        :param frame:
            Frame to get the period of.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Period of the given status frame.

        '''
        assert isinstance(frame, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError

    def setControlFramePeriod(self, frame: int, periodMs: int) -> int:
        '''Sets the period of the given control frame.

        :param frame:
            Frame whose period is to be changed.
        :param periodMs:
            Period in ms for the given frame.
        :returns: Error Code generated by function. 0 indicates no error.

        '''
        assert isinstance(frame, int)
        assert isinstance(periodMs, int)
        raise NotImplementedError





import tkinter as tk
from pyfrc.sim.ui_widgets import CheckButtonWrapper, Tooltip, ValueWidget

from hal_impl.data import hal_data
from .constants import TalonSRXConst as tsrxc

#from hal import TalonSRXConst as tsrxc

class CtreUI:
    
    can_mode_map = {
         tsrxc.kMode_CurrentCloseLoop: 'PercentVbus',
         tsrxc.kMode_DutyCycle:'PercentVbus',
         tsrxc.kMode_NoDrive:'Disabled',
         tsrxc.kMode_PositionCloseLoop:'Position',
         tsrxc.kMode_SlaveFollower:'Follower',
         tsrxc.kMode_VelocityCloseLoop:'Speed',
         tsrxc.kMode_VoltCompen:'Voltage',
         tsrxc.kMode_MotionMagic:'Motion Magic'
    }
    
    def __init__(self):
        self.can = {}
    
    def update_tk_widgets(self, sim):
        
        for k, data in hal_data['CAN'].items():
            
            if data['type'] != 'talonsrx':
                continue
            
            if not data['sim_display']:
                self._add_CAN(sim, k, data)
                data['sim_display'] = True
           
            (motor, fl, rl, mode_lbl_txt, enc_txt, analog_txt, pwm_txt) = self.can[k]
            data = hal_data['CAN'][k]
            mode = data['mode_select']
            mode_lbl_txt.set(self.can_mode_map[mode])
            #change how output works based on control mode
            if   mode == tsrxc.kMode_DutyCycle :  
                #based on the fact that the vbus has 1023 steps
                motor.set_value(data['value']/1023)
                
            elif mode == tsrxc.kMode_VoltCompen:
                #assume voltage is 12 divide by muliplier in cantalon code (256)
                motor.set_value(data['value']/12/256)
                
            elif mode == tsrxc.kMode_SlaveFollower:
                #follow the value of the motor value is equal too
                try:
                    followed = self.can[data['value']]
                except KeyError:
                    value = 0
                else:
                    value = followed[0].get_value()
                
                motor.set_value(value)
            #
            # currently other control modes are not correctly implemented
            #
            else:
                motor.set_value(data['value'])
                
            enc_txt.set('E: %s' % round(data['enc_position']))
            analog_txt.set('A: %s' % round(data['analog_in_position']))
            pwm_txt.set('P: %s' % data['pulse_width_position'])
            
            ret = fl.sync_value(data['limit_switch_closed_for'])
            if ret is not None:
                data['limit_switch_closed_for'] = ret
                
            ret = rl.sync_value(data['limit_switch_closed_rev'])
            if ret is not None:
                data['limit_switch_closed_rev'] = ret 
    
    def _add_CAN(self, sim, canId, device):
        
        # TODO: this is not flexible
        
        row = len(self.can)*2
        
        lbl = tk.Label(sim.can_slot, text=str(canId))
        lbl.grid(column=0, row=row)
        
        motor = ValueWidget(sim.can_slot, default=0.0)
        motor.grid(column=1, row=row)
        sim.set_tooltip(motor, 'CAN', canId)
        
        fl = CheckButtonWrapper(sim.can_slot, text='F')
        fl.grid(column=2, row=row)
        
        rl = CheckButtonWrapper(sim.can_slot, text='R')
        rl.grid(column=3, row=row)
        
        Tooltip.create(fl, 'Forward limit switch')
        Tooltip.create(rl, 'Reverse limit switch')
        
        mode_lbl_txt = tk.StringVar(value = self.can_mode_map[device['mode_select']])
        mode_label = tk.Label(sim.can_slot, textvariable=mode_lbl_txt)
        mode_label.grid(column=4, row=row)
        
        labels = tk.Frame(sim.can_slot)
        labels.grid(column=0, row=row+1, columnspan=6)
        
        enc_value = tk.StringVar(value='E: 0')
        enc_label = tk.Label(labels, textvariable=enc_value)
        enc_label.pack(side=tk.LEFT)
        
        analog_value = tk.StringVar(value='A: 0')
        analog_label = tk.Label(labels, textvariable=analog_value)
        analog_label.pack(side=tk.LEFT)
        
        pwm_value = tk.StringVar(value='P: 0')
        pwm_label = tk.Label(labels, textvariable=pwm_value)
        pwm_label.pack(side=tk.LEFT)
        
        Tooltip.create(enc_label, "Encoder Input")
        Tooltip.create(analog_label, "Analog Input")
        Tooltip.create(pwm_label, "PWM Input")
        
        self.can[canId] = (motor, fl, rl, mode_lbl_txt, enc_value, analog_value, pwm_value)
        
        



# validated: 2018-01-06 EN d090f9cdc6a1 libraries/driver/include/ctre/phoenix/MotorControl/Faults.h


__all__ = ['FaultsBase']


class FaultsBase:
    fields = []

    def __init__(self, bits = 0):
        mask = 1
        for field in self.fields:
            setattr(self, field, bool(bits & mask))
            mask <<= 1

    def toBitfield(self):
        retval = 0
        mask = 1
        for field in self.fields:
            if getattr(self, field):
                retval |= mask
                mask <<= 1
        return retval

    def hasAnyFault(self):
        return any([getattr(self, field) for field in self.fields])

    def __str__(self):
        return " ".join(["%s:%s" % (field, int(getattr(self,field))) for field in self.fields])





# validated: 2018-01-05 EN 8f8595a20145 java/src/com/ctre/phoenix/CANifier.java
#----------------------------------------------------------------------------
#  Software License Agreement
#
# Copyright (C) Cross The Road Electronics.  All rights
# reserved.
# 
# Cross The Road Electronics (CTRE) licenses to you the right to 
# use, publish, and distribute copies of CRF (Cross The Road) firmware files (*.crf) and Software
# API Libraries ONLY when in use with Cross The Road Electronics hardware products.
# 
# THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT
# LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL
# CROSS THE ROAD ELECTRONICS BE LIABLE FOR ANY INCIDENTAL, SPECIAL, 
# INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF
# PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
# BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE
# THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER
# SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT
# (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE
#----------------------------------------------------------------------------
import enum
import hal

from ._impl import (
    GeneralPin,
    ParamEnum,
    CANifierStatusFrame,
    CANifierControlFrame,
    CANifierFaults,
    CANifierStickyFaults,
    CANifier as CANifierImpl,
)


__all__ = ['CANifier', 'LEDChannel', 'PWMChannel']


class LEDChannel(enum.IntEnum):
    """Enum for the LED Output Channels"""
    LEDChannelA = 0
    LEDChannelB = 1
    LEDChannelC = 2


class PWMChannel(enum.IntEnum):
    """Enum for the PWM Input Channels"""
    PWMChannel0 = 0
    PWMChannel1 = 1
    PWMChannel2 = 2
    PWMChannel3 = 3


class PinValues:
    """Class to hold the pin values."""
    def __init__(self):
        self.QUAD_IDX = False
        self.QUAD_B = False
        self.QUAD_A = False
        self.LIMR = False
        self.LIMF = False
        self.SDA = False
        self.SCL = False
        self.SPI_CS_PWM3 = False
        self.SPI_MISO_PWM2 = False
        self.SPI_MOSI_PWM1 = False
        self.SPI_CLK_PWM0 = False


class CANifier(CANifierImpl):
    """CTRE CANifier
        
    Device for interfacing common devices to the CAN bus.
    """

    PWMChannelCount = 4

    def __init__(self, deviceId: int):
        """
        Constructor.

        :param deviceId: The CAN Device ID of the CANifier.
        """
        super().__init__()
        self.create1(deviceId)
        self.tempPins = [False] * 11
        hal.report(hal.UsageReporting.kResourceType_CANifier, deviceId + 1)

    def setLEDOutput(self, percentOutput: float, ledChannel: LEDChannel):
        """
        Sets the LED Output

        :param percentOutput: Output duty cycle expressed as percentage.
        :param ledChannel: Channel to set the output of.
        """
        if percentOutput > 1:
            percentOutput = 1
        if percentOutput < 0:
            percentOutput = 0
        dutyCycle = int(percentOutput * 1023) # [0,1023]

        super().setLEDOutput(dutyCycle, ledChannel)

    def getGeneralInputs(self, allPins: PinValues):
        """
        Gets the state of all General Pins

        :param allPins: A structure to fill with the current state of all pins.
        """
        _, self.tempPins = super().getGeneralInputs()
        allPins.LIMF = self.tempPins[GeneralPin.LIMF]
        allPins.LIMR = self.tempPins[GeneralPin.LIMR]
        allPins.QUAD_A = self.tempPins[GeneralPin.QUAD_A]
        allPins.QUAD_B = self.tempPins[GeneralPin.QUAD_B]
        allPins.QUAD_IDX = self.tempPins[GeneralPin.QUAD_IDX]
        allPins.SCL = self.tempPins[GeneralPin.SCL]
        allPins.SDA = self.tempPins[GeneralPin.SDA]
        allPins.SPI_CLK_PWM0 = self.tempPins[GeneralPin.SPI_CLK_PWM0P]
        allPins.SPI_MOSI_PWM1 = self.tempPins[GeneralPin.SPI_MOSI_PWM1P]
        allPins.SPI_MISO_PWM2 = self.tempPins[GeneralPin.SPI_MISO_PWM2P]
        allPins.SPI_CS_PWM3 = self.tempPins[GeneralPin.SPI_CS]

    def setPWMOutput(self, pwmChannel: int, dutyCycle: float):
        """
        Sets the PWM Output
        Currently supports PWM 0, PWM 1, and PWM 2

        :param pwmChannel: Index of the PWM channel to output.
        :param dutyCycle: Duty Cycle (0 to 1) to output.  Default period of the signal is 4.2 ms.
        """
        dutyCycle = max(dutyCycle, 0)
        dutyCycle = min(dutyCycle, 1)
        pwmChannel = max(pwmChannel, 0)

        dutyCyc10bit = int(1023 * dutyCycle)

        super().setPWMOutput(pwmChannel, dutyCyc10bit)

    def enablePWMOutput(self, pwmChannel: int, bEnable: bool):
        """
        Enables PWM Outputs
        Currently supports PWM 0, PWM 1, and PWM 2

        :param pwmChannel: Index of the PWM channel to enable.
        :param bEnable: True" enables output on the pwm channel.
        """
        pwmChannel = max(pwmChannel, 0)
        super().enablePWMOutput(pwmChannel, bEnable)

    def getFaults(self, toFill: CANifierFaults):
        """
        Gets the CANifier fault status
        
        :param toFill:
            Container for fault statuses.
        :returns: Error Code generated by function. 0 indicates no error.
        """
        _, bits = super().getFaults()
        toFill.update(bits)
        return self.getLastError()

    def getStickyFaults(self, toFill: CANifierStickyFaults):
        """
        Gets the CANifier sticky fault status

        :param toFill:
            Container for sticky fault statuses.
        :returns: Error Code generated by function. 0 indicates no error.
        """
        _, bits = super().getStickyFaults()
        toFill.update(bits)
        return self.getLastError()



import enum

class ErrorCode(enum.IntEnum):
    OK = 0
    OKAY = 0
    CAN_MSG_STALE = 1
    CAN_TX_FULL = -1
    TxFailed = -1
    InvalidParamValue = -2
    CAN_INVALID_PARAM = -2
    RxTimeout = -3
    CAN_MSG_NOT_FOUND = -3
    TxTimeout = -4
    CAN_NO_MORE_TX_JOBS = -4
    UnexpectedArbId = -5
    CAN_NO_SESSIONS_AVAIL = -5
    BufferFull = +6
    CAN_OVERFLOW = -6
    SensorNotPresent = -7
    FirmwareTooOld = -8
    CouldNotChangePeriod = -9
    GeneralError = -100
    GENERAL_ERROR = -100
    SIG_NOT_UPDATED = -200
    SigNotUpdated = -200
    NotAllPIDValuesUpdated = -201
    GEN_PORT_ERROR = -300
    PORT_MODULE_TYPE_MISMATCH = -301
    GEN_MODULE_ERROR = -400
    MODULE_NOT_INIT_SET_ERROR = -401
    MODULE_NOT_INIT_GET_ERROR = -402
    WheelRadiusTooSmall = -500
    TicksPerRevZero = -501
    DistanceBetweenWheelsTooSmall = -502
    GainsAreNotSet = -503
    IncompatibleMode = -600
    InvalidHandle = -601
    PulseWidthSensorNotPresent = 10
    GeneralWarning = 100
    FeatureNotSupported = 101
    NotImplemented = 102
    FirmVersionCouldNotBeRetrieved = 103
    FeaturesNotAvailableYet = 104
    ControlModeNotValid = 105
    ControlModeNotSupportedYet = 106
    CascadedPIDNotSupporteYet = 107
    RemoteSensorsNotSupportedYet = 108

class ParamEnum(enum.IntEnum):
    eOnBoot_BrakeMode = 31
    eQuadFilterEn = 91
    eQuadIdxPolarity = 108
    eClearPositionOnIdx = 100
    eMotionProfileHasUnderrunErr = 119
    eClearPosOnLimitF = 144
    eClearPosOnLimitR = 145
    eStatusFramePeriod = 300
    eOpenloopRamp = 301
    eClosedloopRamp = 302
    eNeutralDeadband = 303
    ePeakPosOutput = 305
    eNominalPosOutput = 306
    ePeakNegOutput = 307
    eNominalNegOutput = 308
    eProfileParamSlot_P = 310
    eProfileParamSlot_I = 311
    eProfileParamSlot_D = 312
    eProfileParamSlot_F = 313
    eProfileParamSlot_IZone = 314
    eProfileParamSlot_AllowableErr = 315
    eProfileParamSlot_MaxIAccum = 316
    eClearPositionOnLimitF = 320
    eClearPositionOnLimitR = 321
    eClearPositionOnQuadIdx = 322
    eSampleVelocityPeriod = 325
    eSampleVelocityWindow = 326
    eFeedbackSensorType = 330
    eSelectedSensorPosition = 331
    eFeedbackNotContinuous = 332
    eRemoteSensorSource = 333
    eRemoteSensorDeviceID = 334
    eSensorTerm = 335
    eRemoteSensorClosedLoopDisableNeutralOnLOS = 336
    eForwardSoftLimitThreshold = 340
    eReverseSoftLimitThreshold = 341
    eForwardSoftLimitEnable = 342
    eReverseSoftLimitEnable = 343
    eNominalBatteryVoltage = 350
    eBatteryVoltageFilterSize = 351
    eContinuousCurrentLimitAmps = 360
    ePeakCurrentLimitMs = 361
    ePeakCurrentLimitAmps = 362
    eClosedLoopIAccum = 370
    eCustomParam = 380
    eStickyFaults = 390
    eAnalogPosition = 400
    eQuadraturePosition = 401
    ePulseWidthPosition = 402
    eMotMag_Accel = 410
    eMotMag_VelCruise = 411
    eLimitSwitchSource = 421
    eLimitSwitchNormClosedAndDis = 422
    eLimitSwitchDisableNeutralOnLOS = 423
    eLimitSwitchRemoteDevID = 424
    eYawOffset = 160
    eCompassOffset = 161
    eBetaGain = 162
    eEnableCompassFusion = 163
    eGyroNoMotionCal = 164
    eEnterCalibration = 165
    eFusedHeadingOffset = 166
    eStatusFrameRate = 169
    eAccumZ = 170
    eTempCompDisable = 171
    eMotionMeas_tap_threshX = 172
    eMotionMeas_tap_threshY = 173
    eMotionMeas_tap_threshZ = 174
    eMotionMeas_tap_count = 175
    eMotionMeas_tap_time = 176
    eMotionMeas_tap_time_multi = 177
    eMotionMeas_shake_reject_thresh = 178
    eMotionMeas_shake_reject_time = 179
    eMotionMeas_shake_reject_timeout = 180

class CANifierControlFrame(enum.IntEnum):
    CANifier_Control_1_General = 50593792
    CANifier_Control_2_PwmOutput = 50593856

class CANifierStatusFrame(enum.IntEnum):
    CANifierStatusFrame_Status_1_General = 267264
    CANifierStatusFrame_Status_2_General = 267328
    CANifierStatusFrame_Status_3_PwmInputs0 = 267392
    CANifierStatusFrame_Status_4_PwmInputs1 = 267456
    CANifierStatusFrame_Status_5_PwmInputs2 = 267520
    CANifierStatusFrame_Status_6_PwmInputs3 = 267584
    CANifierStatusFrame_Status_8_Misc = 267712

class SetValueMotionProfile(enum.IntEnum):
    Disable = 0
    Enable = 1
    Hold = 2

class ControlFrame(enum.IntEnum):
    Control_3_General = 262272
    Control_4_Advanced = 262336
    Control_6_MotProfAddTrajPoint = 262464
class ControlFrameEnhanced(enum.IntEnum):
    Control_3_General_ = 262272
    Control_4_Advanced_ = 262336
    Control_5_FeedbackOutputOverride_ = 262400
    Control_6_MotProfAddTrajPoint_ = 262464

class ControlMode(enum.IntEnum):
    PercentOutput = 0
    Position = 1
    Velocity = 2
    Current = 3
    Follower = 5
    MotionProfile = 6
    MotionMagic = 7
    MotionMagicArc = 8
    MotionProfileArc = 10
    Disabled = 15

class FeedbackDevice(enum.IntEnum):
    None_ = -1
    QuadEncoder = 0
    Analog = 2
    Tachometer = 4
    PulseWidthEncodedPosition = 8
    SensorSum = 9
    SensorDifference = 10
    RemoteSensor0 = 11
    RemoteSensor1 = 12
    SoftwareEmulatedSensor = 15
    CTRE_MagEncoder_Absolute = PulseWidthEncodedPosition
    CTRE_MagEncoder_Relative = QuadEncoder
class RemoteFeedbackDevice(enum.IntEnum):
    RemoteFeedbackDevice_None = -1
    RemoteFeedbackDevice_SensorSum = 9
    RemoteFeedbackDevice_SensorDifference = 10
    RemoteFeedbackDevice_RemoteSensor0 = 11
    RemoteFeedbackDevice_RemoteSensor1 = 12
    RemoteFeedbackDevice_SoftwareEmulatedSensor = 15

class LimitSwitchSource(enum.IntEnum):
    LimitSwitchSource_FeedbackConnector = 0
    LimitSwitchSource_RemoteTalonSRX = 1
    LimitSwitchSource_RemoteCANifier = 2
    LimitSwitchSource_Deactivated = 3
class RemoteLimitSwitchSource(enum.IntEnum):
    RemoteLimitSwitchSource_RemoteTalonSRX = 1
    RemoteLimitSwitchSource_RemoteCANifier = 2
    RemoteLimitSwitchSource_Deactivated = 3
class LimitSwitchNormal(enum.IntEnum):
    LimitSwitchNormal_NormallyOpen = 0
    LimitSwitchNormal_NormallyClosed = 1
    LimitSwitchNormal_Disabled = 2

class NeutralMode(enum.IntEnum):
    EEPROMSetting = 0
    Coast = 1
    Brake = 2

class RemoteSensorSource(enum.IntEnum):
    RemoteSensorSource_Off = 0
    RemoteSensorSource_TalonSRX_SelectedSensor = 1
    RemoteSensorSource_Pigeon_Yaw = 2
    RemoteSensorSource_Pigeon_Pitch = 3
    RemoteSensorSource_Pigeon_Roll = 4
    RemoteSensorSource_CANifier_Quadrature = 5
    RemoteSensorSource_CANifier_PWMInput0 = 6
    RemoteSensorSource_CANifier_PWMInput1 = 7
    RemoteSensorSource_CANifier_PWMInput2 = 8
    RemoteSensorSource_CANifier_PWMInput3 = 9

class SensorTerm(enum.IntEnum):
    SensorTerm_Sum0 = 0
    SensorTerm_Sum1 = 1
    SensorTerm_Diff0 = 2
    SensorTerm_Diff1 = 3

class StatusFrameEnhanced(enum.IntEnum):
    Status_1_General = 5120
    Status_2_Feedback0 = 5184
    Status_4_AinTempVbat = 5312
    Status_6_Misc = 5440
    Status_7_CommStatus = 5504
    Status_9_MotProfBuffer = 5632
    Status_10_MotionMagic = 5696
    Status_12_Feedback1 = 5824
    Status_13_Base_PIDF0 = 5888
    Status_14_Turn_PIDF1 = 5952
    Status_15_FirmareApiStatus = 6016
    Status_3_Quadrature = 5248
    Status_8_PulseWidth = 5568
    Status_11_UartGadgeteer = 5760
class StatusFrame(enum.IntEnum):
    Status_1_General_ = 5120
    Status_2_Feedback0_ = 5184
    Status_4_AinTempVbat_ = 5312
    Status_6_Misc_ = 5440
    Status_7_CommStatus_ = 5504
    Status_9_MotProfBuffer_ = 5632
    Status_10_MotionMagic_ = 5696
    Status_12_Feedback1_ = 5824
    Status_13_Base_PIDF0_ = 5888
    Status_14_Turn_PIDF1_ = 5952
    Status_15_FirmareApiStatus_ = 6016

class VelocityMeasPeriod(enum.IntEnum):
    Period_1Ms = 1
    Period_2Ms = 2
    Period_5Ms = 5
    Period_10Ms = 10
    Period_20Ms = 20
    Period_25Ms = 25
    Period_50Ms = 50
    Period_100Ms = 100

class PigeonIMU_ControlFrame(enum.IntEnum):
    PigeonIMU_CondStatus_Control_1 = 272384

class PigeonIMU_StatusFrame(enum.IntEnum):
    PigeonIMU_CondStatus_1_General = 270336
    PigeonIMU_CondStatus_9_SixDeg_YPR = 270848
    PigeonIMU_CondStatus_6_SensorFusion = 270656
    PigeonIMU_CondStatus_11_GyroAccum = 270976
    PigeonIMU_CondStatus_2_GeneralCompass = 270400
    PigeonIMU_CondStatus_3_GeneralAccel = 270464
    PigeonIMU_CondStatus_10_SixDeg_Quat = 270912
    PigeonIMU_RawStatus_4_Mag = 269504
    PigeonIMU_BiasedStatus_2_Gyro = 269376
    PigeonIMU_BiasedStatus_4_Mag = 269504
    PigeonIMU_BiasedStatus_6_Accel = 269632




# validated: 2018-01-05 EN d376a57d7f94 java/src/com/ctre/phoenix/motorcontrol/can/TalonSRX.java
#----------------------------------------------------------------------------
#  Software License Agreement
#
# Copyright (C) Cross The Road Electronics.  All rights
# reserved.
# 
# Cross The Road Electronics (CTRE) licenses to you the right to 
# use, publish, and distribute copies of CRF (Cross The Road) firmware files (*.crf) and Software
# API Libraries ONLY when in use with Cross The Road Electronics hardware products.
# 
# THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT
# LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL
# CROSS THE ROAD ELECTRONICS BE LIABLE FOR ANY INCIDENTAL, SPECIAL, 
# INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF
# PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
# BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE
# THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER
# SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT
# (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE
#----------------------------------------------------------------------------
import hal
from .basemotorcontroller import BaseMotorController
from ._impl import MotController


__all__ = ['TalonSRX']


class TalonSRX(BaseMotorController):
    """CTRE Talon SRX Motor Controller when used on CAN Bus.
    
    Don't use this directly, use :class:`.WPI_TalonSRX`"""

    def __init__(self, deviceNumber):
        super().__init__(deviceNumber | 0x02040000)
        hal.report(hal.UsageReporting.kResourceType_CANTalonSRX, deviceNumber + 1)

    def configForwardLimitSwitchSource(self, type, normalOpenOrClose, timeoutMs: int):
        """
        Configures a limit switch for a local/remote source.

        For example, a CAN motor controller may need to monitor the Limit-R pin
        of another Talon, CANifier, or local Gadgeteer feedback connector.

        If the sensor is remote, a device ID of zero is assumed.
        If that's not desired, use the four parameter version of this function.

        :param type:
            Limit switch source.
            User can choose between the feedback connector, remote Talon SRX, CANifier, or deactivate the feature.
        :param normalOpenOrClose:
            Setting for normally open, normally closed, or disabled. This setting
            matches the web-based configuration drop down.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.
        """
        return super().configForwardLimitSwitchSource(type, normalOpenOrClose, 0x00000000, timeoutMs)

    def configReverseLimitSwitchSource(self, type, normalOpenOrClose, timeoutMs: int):
        """
        Configures a limit switch for a local/remote source.
        
        For example, a CAN motor controller may need to monitor the Limit-R pin
        of another Talon, CANifier, or local Gadgeteer feedback connector.
        
        If the sensor is remote, a device ID of zero is assumed. If that's not
        desired, use the four parameter version of this function.
        
        :param type:
            Limit switch source. See :class:`.LimitSwitchSource` User can choose
            between the feedback connector, remote Talon SRX, CANifier, or
            deactivate the feature.
        :param normalOpenOrClose:
            Setting for normally open, normally closed, or disabled. This
            setting matches the web-based configuration drop down.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.
        """
        return super().configReverseLimitSwitchSource(type, normalOpenOrClose, 0x00000000, timeoutMs)




import re

_annotations = {
    'short': int,
    'int': 'int',
    'uint32_t': 'int',
    'double': 'float',
    'char': 'str',
    'bool': 'bool',
    'ctre::phoenix::ErrorCode': 'int',
}

def _gen_check(pname, ptype):

    # TODO: This does checks on normal types, but if you pass a ctypes value
    #       in then this does not check those properly.

    if ptype == 'bool':
        return 'isinstance(%s, bool)' % pname

    elif ptype in ['float', 'double']:
        return 'isinstance(%s, (int, float))' % pname

    #elif ptype is C.c_char:
    #    return 'isinstance(%s, bytes) and len(%s) == 1' % (pname, pname)
    #elif ptype is C.c_wchar:
    #    return 'isinstance(%s, str) and len(%s) == 1' % (pname, pname)
    #elif ptype is C.c_char_p:
    #    return "%s is None or isinstance(%s, bytes) or getattr(%s, '_type_') is _C.c_char" % (pname, pname, pname)
    #elif ptype is C.c_wchar_p:
    #    return '%s is None or isinstance(%s, bytes)' % (pname, pname)

    elif ptype in ['int', 'long']:
        return 'isinstance(%s, int)' % pname
    #elif ptype in [C.c_byte, C.c_int8]:
    #    return 'isinstance(%s, int) and %s < %d and %s > -%d' % (pname, pname, 1<<7, pname, 1<<7)
    elif ptype in ['short', 'int16_t']:
        return 'isinstance(%s, int) and %s < %d and %s > -%d' % (pname, pname, 1<<15, pname, 1<<15)
    elif ptype == 'int32_t':
        return 'isinstance(%s, int) and %s < %d and %s > -%d' % (pname, pname, 1<<31, pname, 1<<31)
    elif ptype == 'int64_t':
        return 'isinstance(%s, int) and %s < %d and %s > -%d' % (pname, pname, 1<<63, pname, 1<<63)

    elif ptype == 'size_t':
        return 'isinstance(%s, int)' % (pname)
    elif ptype == 'uint8_t':
        return 'isinstance(%s, int) and %s < %d and %s >= 0' % (pname, pname, 1<<8, pname)
    elif ptype == 'uint16_t':
        return 'isinstance(%s, int) and %s < %d and %s >= 0' % (pname, pname, 1<<16, pname)
    elif ptype == 'uint32_t':
        return 'isinstance(%s, int) and %s < %d and %s >= 0' % (pname, pname, 1<<32, pname)
    elif ptype == 'uint64_t':
        return 'isinstance(%s, int) and %s < %d and %s >= 0' % (pname, pname, 1<<64, pname)

    elif ptype is None:
        return '%s is None' % pname

    else:
        # TODO: do validation here
        #return 'isinstance(%s, %s)' % (pname, type(ptype).__name__)
        return None


def _to_annotation(ctypename):
    return _annotations[ctypename]

def header_hook(header, data):
    '''Called for each header'''

    # fix enum names
    for e in header.enums:
        ename = e['name'].split('_')[0] + '_'
        for v in e['values']:
            name = v['name']
            if name.startswith(ename):
                name = name[len(ename):]
            if name == 'None':
                name = 'NONE'
            v['x_name'] = name

def function_hook(fn, data):
    '''Called for each function in the header'''

    # Mangle the name appropriately
    m = re.match(r'c_%s_(.*)' % data['module_name'], fn['name'])
    if not m:
        raise Exception("Unexpected fn %s" % fn['name'])

    x_name = m.group(1)
    x_name = x_name[0].lower() + x_name[1:]

    # Setup the data for wrapping the CCI interface

    returns = fn['returns']

    x_in_params = []
    x_out_params = []

    # returns can be an out param
    # -> but in a check, it's not

    # final_out is really what we want?

    # parameter names, types

    # inner call

    # return

    x_ret_names = []
    x_ret_types = []

    x_param_checks = []

    param_offset = 0 if x_name.startswith('create') else 1

    for i, p in enumerate(fn['parameters'][param_offset:]):
        if p['name'] == '':
            p['name'] = 'param%s' % i
        p['x_type'] = p['raw_type']
        p['x_callname'] = p['name']

        # Python annotations for sim
        p['x_pyann_type'] = _to_annotation(p['raw_type'])
        p['x_pyann'] = '%(name)s: %(x_pyann_type)s' % p

        if p['pointer']:
            p['x_callname'] = '&%(x_callname)s' % p
            x_out_params.append(p)
        elif p['array']:
            asz = p.get('array_size', 0)
            if asz:
                p['x_pyann_type'] = 'list'
                p['x_type'] = 'std::array<%s, %s>' % (p['x_type'], asz)
                p['x_callname'] = '%(x_callname)s.data()' % p
            else:
                # it's a vector
                pass

            x_out_params.append(p)
        else:
            chk = _gen_check(p['name'], p['raw_type'])
            if chk:
                x_param_checks.append('assert %s' % chk)
            x_in_params.append(p)


        p['x_decl'] = '%s %s' % (p['x_type'], p['name'])

    # if there are out params, then change returns to be a tuple
    if x_out_params:
        p = ', '.join([p['x_type'] for p in x_out_params])
        returns = 'std::tuple<%s, %s>' % (fn['returns'], p)

    x_callend = ''
    x_return = ''

    if returns == 'void':
        x_callstart = ''
        x_pyann_ret = 'None'
    else:
        #if returns == 'ctre::phoenix::ErrorCode':
        #    x_callstart = 'auto __ret = CheckCTRCode('
        #    x_callend = ')'
        #else:

        # TODO: should we throw on a CTRE error? Their Java APIs don't,
        #       so we probably shouldn't either
        x_callstart = 'auto __ret = '

        if x_out_params:
            x_return = 'return std::make_tuple(__ret, %s);'
            x_return %= ', '.join([p['name'] for p in x_out_params])

            x_pyann_ret = 'typing.Tuple[%s, %s]'
            x_pyann_ret %= (_to_annotation(fn['returns']),
                            ', '.join([p['x_pyann_type'] for p in x_out_params]))
        else:
            x_return = 'return __ret;'
            x_pyann_ret = _to_annotation(returns)

    # Temporary values to store out parameters in
    x_temprefs = ''
    if x_out_params:
        x_temprefs = \
            ';'.join([
                '%(x_type)s %(name)s' % p for p in x_out_params
            ]) + ';'

    py_self_comma = ', ' if x_in_params else ''

    # x_call_params

    data = data.get('data', {}).get(fn['name'])

    # lazy :)
    fn.update(locals())



# validated: 2018-01-06 EN d090f9cdc6a1 libraries/driver/include/ctre/phoenix/MotorControl/StickyFaults.h
from .faultsbase import FaultsBase

__all__ = ['StickyFaults']


class StickyFaults(FaultsBase):
    fields = ['underVoltage', 'forwardLimitSwitch', 'reverseLimitSwitch', 'forwardSoftLimit', 'reverseSoftLimit',
              'resetDuringEn', 'sensorOverflow', 'sensorOutOfPhase', 'hardwareESDReset', 'remoteLossOfSignal']


import hal
from collections import namedtuple
from .faults import Faults
from .stickyfaults import StickyFaults
from .canifierfaults import CANifierFaults
from .canifierstickyfaults import CANifierStickyFaults
from .pigeonfaults import PigeonIMU_Faults
from .pigeonstickyfaults import PigeonIMU_StickyFaults
from .trajectorypoint import TrajectoryPoint
from .motionprofilestatus import MotionProfileStatus

if hal.isSimulation():
    from .autogen.canifier_sim import CANifier
    from .autogen.motcontroller_sim import MotController
    from .autogen.pigeonimu_sim import PigeonIMU
    from .autogen.ctre_sim_enums import (
        ErrorCode,
        NeutralMode,
        RemoteFeedbackDevice,
        FeedbackDevice,
        StatusFrame,
        StatusFrameEnhanced,
        VelocityMeasPeriod,
        RemoteLimitSwitchSource,
        LimitSwitchNormal,
        LimitSwitchSource,
        ParamEnum,
        CANifierStatusFrame,
        CANifierControlFrame,
        ParamEnum,
        PigeonIMU_StatusFrame,
        PigeonIMU_ControlFrame,
    )
else:
    from .ctre_roborio import (
        CANifier,
        MotController,
        PigeonIMU
    )
    from .ctre_roborio import (
        ErrorCode,
        NeutralMode,
        RemoteFeedbackDevice,
        FeedbackDevice,
        StatusFrame,
        StatusFrameEnhanced,
        VelocityMeasPeriod,
        RemoteLimitSwitchSource,
        LimitSwitchNormal,
        LimitSwitchSource,
        ParamEnum,
        CANifierStatusFrame,
        CANifierControlFrame,
        ParamEnum,
        PigeonIMU_StatusFrame,
        PigeonIMU_ControlFrame,
    )


# todo: eliminate
import enum
from unittest.mock import MagicMock

class ControlMode (enum.IntEnum):
    PercentOutput = 0
    Position = 1
    Velocity = 2
    Current = 3
    Follower = 5
    MotionProfile = 6
    MotionMagic = 7
    MotionMagicArc = 8
    #TimedPercentOutput = 9
    MotionProfileArc = 10
    Disabled = 15


GeneralPin = MagicMock()



# validated: 2018-01-06 EN d090f9cdc6a1 libraries/driver/include/ctre/phoenix/CANifierStickyFaults.h
from .faultsbase import FaultsBase

__all__ = ['CANifierStickyFaults']


class CANifierStickyFaults(FaultsBase):
    pass



# validated: 2018-01-06 EN d090f9cdc6a1 java/src/com/ctre/phoenix/motorcontrol/can/BaseMotorController.java
#----------------------------------------------------------------------------
#  Software License Agreement
#
# Copyright (C) Cross The Road Electronics.  All rights
# reserved.
# 
# Cross The Road Electronics (CTRE) licenses to you the right to 
# use, publish, and distribute copies of CRF (Cross The Road) firmware files (*.crf) and Software
# API Libraries ONLY when in use with Cross The Road Electronics hardware products.
# 
# THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT
# LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL
# CROSS THE ROAD ELECTRONICS BE LIABLE FOR ANY INCIDENTAL, SPECIAL, 
# INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF
# PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
# BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE
# THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER
# SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT
# (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE
#----------------------------------------------------------------------------
import typing
from .sensorcollection import SensorCollection
from ._impl import (
    MotController,
    ControlMode,
    NeutralMode,
    FeedbackDevice,
    StatusFrame,
    StatusFrameEnhanced,
    VelocityMeasPeriod,
    RemoteLimitSwitchSource,
    LimitSwitchNormal,
    LimitSwitchSource,
    TrajectoryPoint,
    MotionProfileStatus,
    Faults,
    StickyFaults,
    ParamEnum,
)


__all__ = ["BaseMotorController"]


class BaseMotorController(MotController):
    """Base motor controller features for all CTRE CAN motor controllers."""

    def __init__(self, arbId: int):
        """
        Constructor for motor controllers.
        
        :param arbId:
        """
        super().__init__()

        self.create1(arbId)
        self.arbId = arbId
        self.sensorColl = SensorCollection(self)
        self.motionProfStats = [0] * 9
        self.controlMode = ControlMode.PercentOutput
        self.sendMode = ControlMode.PercentOutput

    def getDeviceID(self):
        """
        Returns the Device ID
        
        :returns: Device number.
        """
        return self.getDeviceNumber()[1]

    def set(self, mode: ControlMode, demand0: float, demand1: float = 0.0):
        """
        Sets the appropriate output on the talon, depending on the mode.

        :param mode:
            The output mode to apply.
        :param demand0: 
            The output value to apply. such as advanced feed forward and/or cascaded close-looping in firmware.

            In :attr:`.ControlMode.PercentOutput`, the output is between -1.0 and 1.0, with 0.0 as
            stopped. 
            
            In :attr:`.ControlMode.Voltage` mode, output value is in volts. 
            
            In :attr:`.ControlMode.Current` mode, output value is in amperes. 
            
            In :attr:`.ControlMode.Speed` mode, output value is in position change / 100ms. 
            
            In :attr:`.ControlMode.Position` mode, output value is in encoder ticks or an analog value, depending on the sensor. 
            
            In :attr:`.ControlMode.Follower` mode, the output value is the integer device ID of the talon to duplicate.
        :type demand0: float
        :param demand1:
            Supplemental value.  This will also be control mode specific for future features.
        :type demand1: float

        see :meth:`.selectProfileSlot` to choose between the two sets of gains.
        """
        self.controlMode = mode
        self.sendMode = mode

        if self.controlMode == ControlMode.PercentOutput:
            self.setDemand(self.sendMode, int(1023 * demand0), 0)
        elif self.controlMode == ControlMode.Follower:
            # did caller specify device ID
            if 0 <= demand0 <= 62:
                work = self.getBaseID()
                work >>= 16
                work <<= 8
                work |= int(demand0) & 0xFF
            else:
                work = int(demand0)
            self.setDemand(self.sendMode, work, 0)
        elif self.controlMode in [ControlMode.Velocity, ControlMode.Position, ControlMode.MotionMagic, ControlMode.MotionMagicArc, ControlMode.MotionProfile]:
            self.setDemand(self.sendMode, int(demand0), 0)
        elif self.controlMode == ControlMode.Current:
            self.setDemand(self.sendMode, int(1000. * demand0), 0) # milliamps 
        else:
            self.setDemand(self.sendMode, 0, 0)

    def neutralOutput(self):
        """
        Neutral the motor output by setting control mode to disabled."""
        self.set(ControlMode.Disabled, 0, 0)

    def setInverted(self, invert: bool):
        """
        Inverts the hbridge output of the motor controller.

        This does not impact sensor phase and should not be used to correct sensor polarity.

        This will invert the hbridge output but NOT the LEDs.
        This ensures....

        - Green LEDs always represents positive request from robot-controller/closed-looping mode.
        - Green LEDs correlates to forward limit switch.
        - Green LEDs correlates to forward soft limit.
        
        :param invert:
            Invert state to set.
        """
        self.invert = invert
        super().setInverted(invert)

    def getInverted(self):
        """:returns: invert setting of motor output"""
        return self.invert

    def getMotorOutputVoltage(self):
        """:returns: applied voltage to motor in volts"""
        return self.getBusVoltage() * self.getMotorOutputPercent()

    def getMotionProfileStatus(self):
        """
        Retrieve all status information.
        For best performance, Caller can snapshot all status information regarding the
        motion profile executer.
        
        :param statusToFill: 
            Caller supplied object to fill.
        
            The members are filled, as follows...
            
            topBufferRem:   The available empty slots in the trajectory buffer.
            The robot API holds a "top buffer" of trajectory points, so your applicaion
            can dump several points at once.  The API will then stream them into the
            low-level buffer, allowing the motor controller to act on them.
            
            topBufferRem: The number of points in the top trajectory buffer.
            
            btmBufferCnt: The number of points in the low level controller buffer.
            
            hasUnderrun:    Set if isUnderrun ever gets set.
            Only is cleared by clearMotionProfileHasUnderrun() to ensure
            
            isUnderrun:     This is set if controller needs to shift a point from its buffer into
            the active trajectory point however
            the buffer is empty.
            This gets cleared automatically when is resolved.
            
            activePointValid:   True if the active trajectory point has not empty, false otherwise. The members in activePoint are only valid if this signal is set.
            
            isLast: is set/cleared based on the MP executer's current
            trajectory point's IsLast value.  This assumes
            IsLast was set when PushMotionProfileTrajectory
            was used to insert the currently processed trajectory
            point.
            
            profileSlotSelect: The currently processed trajectory point's
            selected slot.  This can differ in the currently selected slot used
            for Position and Velocity servo modes
            
            outputEnable:       The current output mode of the motion profile
            executer (disabled, enabled, or hold).  When changing the set()
            value in MP mode, it's important to check this signal to
            confirm the change takes effect before interacting with the top buffer.
        :returns: (errorcode, :class:`.MotionProfileStatus`)
        """
        retval, *fields = super().getMotionProfileStatus()
        statusToFill = MotionProfileStatus(*fields)
        return retval, statusToFill

    def pushMotionProfileTrajectory(self, trajPt: TrajectoryPoint):
        """Push another trajectory point into the top level buffer (which is emptied
        into the motor controller's bottom buffer as room allows).

        :param trajPt: to push into buffer.
            The members should be filled in with these values...

            position:  servo position in sensor units.
            velocity:  velocity to feed-forward in sensor units per 100ms.
            profileSlotSelect:  which slot to pull PIDF gains from.  Currently
            supports 0,1,2,3.

            isLastPoint:  set to nonzero to signal motor controller to keep processing this
                trajectory point, instead of jumping to the next one
                when timeDurMs expires.  Otherwise MP executer will
                eventually see an empty buffer after the last point
                expires, causing it to assert the IsUnderRun flag.
                However this may be desired if calling application
                never wants to terminate the MP.
            zeroPos:  set to nonzero to signal motor controller to "zero" the selected
                position sensor before executing this trajectory point.
                Typically the first point should have this set only thus
                allowing the remainder of the MP positions to be relative to
                zero.
        :returns: CTR_OKAY if trajectory point push ok. ErrorCode if buffer is
            full due to kMotionProfileTopBufferCapacity.
        """
        return super().pushMotionProfileTrajectory(trajPt.position, trajPt.velocity, trajPt.headingDeg,
                                                   trajPt.profileSlotSelect, trajPt.isLastPoint, trajPt.zeroPos)

    def getFaults(self) -> typing.Tuple[int, Faults]:
        '''Gets the last error generated by this object. Not all functions return an
        error code but can potentially report errors. This function can be used
        to retrieve those error codes.

        :returns: Last Error Code generated by a function, faults
        '''
        errcode, bits = super().getFaults()
        return errcode, Faults(bits)

    def getStickyFaults(self) -> typing.Tuple[int, StickyFaults]:
        """
        Polls the various sticky fault flags.
        
        :returns: Last Error Code generated by a function, sticky faults
        """
        _, bits = super().getStickyFaults()
        return self.getLastError(), StickyFaults(bits)

    def getBaseID(self) -> int:
        return self.arbId

    def follow(self, masterToFollow: 'BaseMotorController'):
        """
        Set the control mode and output value so that this motor controller will
        follow another motor controller. Currently supports following Victor SPX
        and Talon SRX.
        """
        id32 = masterToFollow.getBaseID()
        id24 = id32
        id24 >>= 16
        id24 = id24 & 0xFFFF
        id24 <<= 8
        id24 |= (id32 & 0xFF)
        self.set(ControlMode.Follower, id24)

    def valueUpdated(self):
        """
        When master makes a device, this routine is called to signal the update."""
        pass
        
    def getSensorCollection(self) -> SensorCollection:
        """
        :returns: object that can get/set individual raw sensor values."""
        return self.sensorColl
        
    def getControlMode(self) -> ControlMode:
        """
        :returns: control mode motor controller is in"""
        return self.controlMode



# validated: 2018-01-05 EN d376a57d7f94 java/src/com/ctre/phoenix/motorcontrol/can/VictorSPX.java
#----------------------------------------------------------------------------
#  Software License Agreement
#
# Copyright (C) Cross The Road Electronics.  All rights
# reserved.
# 
# Cross The Road Electronics (CTRE) licenses to you the right to 
# use, publish, and distribute copies of CRF (Cross The Road) firmware files (*.crf) and Software
# API Libraries ONLY when in use with Cross The Road Electronics hardware products.
# 
# THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT
# LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL
# CROSS THE ROAD ELECTRONICS BE LIABLE FOR ANY INCIDENTAL, SPECIAL, 
# INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF
# PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
# BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE
# THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER
# SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT
# (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE
#----------------------------------------------------------------------------
import hal
from .basemotorcontroller import BaseMotorController


__all__ = ['VictorSPX']


class VictorSPX (BaseMotorController):
    """VEX Victor SPX Motor Controller when used on CAN Bus.
    
    Don't use this directly, use :class:`.WPI_VictorSPX`"""

    def __init__(self, deviceNumber: int):
        """Constructor

        :param deviceNumber:
            [0,62]
        """
        super().__init__(deviceNumber | 0x01040000)
        hal.report(hal.UsageReporting.kResourceType_CTRE_future1, deviceNumber + 1)



# validated: 2018-01-05 EN 2d574139d4e2 java/src/com/ctre/phoenix/motorcontrol/SensorCollection.java
#----------------------------------------------------------------------------
#  Software License Agreement
#
# Copyright (C) Cross The Road Electronics.  All rights
# reserved.
# 
# Cross The Road Electronics (CTRE) licenses to you the right to 
# use, publish, and distribute copies of CRF (Cross The Road) firmware files (*.crf) and Software
# API Libraries ONLY when in use with Cross The Road Electronics hardware products.
# 
# THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT
# LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL
# CROSS THE ROAD ELECTRONICS BE LIABLE FOR ANY INCIDENTAL, SPECIAL, 
# INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF
# PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
# BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE
# THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER
# SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT
# (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE
#----------------------------------------------------------------------------
from ._impl import MotController


__all__ = ["SensorCollection"]


class SensorCollection:

    def __init__(self, motorController):
        """
        :type motorController: :class:`.BaseMotorController`"""
        self.motorController = motorController

    @property
    def impl(self):
        return self.motorController

    @property
    def handle(self):
        return self.motorController.handle

    def getAnalogIn(self) -> int:
        return self.impl.getAnalogIn()

    def setAnalogPosition(self, newPosition: int, timeoutMs: int):
        return self.impl.setAnalogPosition(newPosition, timeoutMs)

    def getAnalogInRaw(self) -> int:
        return self.impl.getAnalogInRaw()

    def getAnalogInVel(self) -> int:
        return self.impl.getAnalogInVel()

    def getQuadraturePosition(self) -> int:
        return self.impl.getQuadraturePosition()

    def setQuadraturePosition(self, newPosition: int, timeoutMs: int):
        return self.impl.setQuadraturePosition(newPosition, timeoutMs)

    def getQuadratureVelocity(self) -> int:
        return self.impl.getQuadratureVelocity()

    def getPulseWidthPosition(self) -> int:
        return self.impl.getPulseWidthPosition()

    def setPulseWidthPosition(self, newPosition: int, timeoutMs: int):
        return self.impl.setPulseWidthPosition(newPosition, timeoutMs)

    def getPulseWidthVelocity(self) -> int:
        return self.impl.getPulseWidthVelocity()

    def getPulseWidthRiseToFallUs(self) -> int:
        return self.impl.getPulseWidthRiseToFallUs()

    def getPulseWidthRiseToRiseUs(self) -> int:
        return self.impl.getPulseWidthRiseToRiseUs()

    def getPinStateQuadA(self) -> bool:
        """
        Gets pin state quad a.

        :returns: the pin state quad a (1 if asserted, 0 if not asserted).
        """
        return self.impl.getPinStateQuadA() != 0

    def getPinStateQuadB(self) -> bool:
        """
        Gets pin state quad b.

        :returns: Digital level of QUADB pin (1 if asserted, 0 if not asserted).
        """
        return self.impl.getPinStateQuadB() != 0

    def getPinStateQuadIdx(self) -> bool:
        """
        Gets pin state quad index.

        :returns: Digital level of QUAD Index pin (1 if asserted, 0 if not asserted).
        """
        return self.impl.getPinStateQuadIdx()[1] != 0

    def isFwdLimitSwitchClosed(self) -> bool:
        """
        Is forward limit switch closed.

        This function works regardless if limit switch feature is
        enabled.

        :returns: '1' iff forward limit switch is closed, 0 iff switch is open.
        """
        return self.impl.isFwdLimitSwitchClosed()[1] != 0

    def isRevLimitSwitchClosed(self) -> bool:
        """
        Is reverse limit switch closed.

        This function works regardless if limit switch feature is
        enabled.

        :returns: '1' iff reverse limit switch is closed, 0 iff switch is open.
        """
        return self.impl.isRevLimitSwitchClosed()[1] != 0




class TalonSRXConst:
    kDefaultControlPeriodMs = 10

    # mode select enumerations
    kMode_DutyCycle = 0
    kMode_PositionCloseLoop = 1
    kMode_VelocityCloseLoop = 2
    kMode_CurrentCloseLoop = 3
    kMode_VoltCompen = 4
    kMode_SlaveFollower = 5
    kMode_MotionProfile = 6
    kMode_MotionMagic = 7
    kMode_NoDrive = 15

    # limit switch enumerations
    kLimitSwitchOverride_UseDefaultsFromFlash = 1
    kLimitSwitchOverride_DisableFwd_DisableRev = 4
    kLimitSwitchOverride_DisableFwd_EnableRev = 5
    kLimitSwitchOverride_EnableFwd_DisableRev = 6
    kLimitSwitchOverride_EnableFwd_EnableRev = 7

    # brake override enumerations
    kBrakeOverride_UseDefaultsFromFlash = 0
    kBrakeOverride_OverrideCoast = 1
    kBrakeOverride_OverrideBrake = 2

    # feedback device enumerations
    kFeedbackDev_DigitalQuadEnc = 0
    kFeedbackDev_AnalogPot = 2
    kFeedbackDev_AnalogEncoder = 3
    kFeedbackDev_CountEveryRisingEdge = 4
    kFeedbackDev_CountEveryFallingEdge = 5
    kFeedbackDev_CtreMagEncoder_Relative = 6
    kFeedbackDev_CtreMagEncoder_Absolute = 7
    kFeedbackDev_PosIsPulseWidth = 8
    
    # feedback device status enumerations
    kFeedbackDevStatus_Unknown = 0
    kFeedbackDevStatus_Present = 1
    kFeedbackDevStatus_NotPresent = 2

    # ProfileSlotSelect enumerations
    kProfileSlotSelect_Slot0 = 0
    kProfileSlotSelect_Slot1 = 1

    # status frame rate types
    kStatusFrame_General = 0
    kStatusFrame_Feedback = 1
    kStatusFrame_Encoder = 2
    kStatusFrame_AnalogTempVbat = 3
    kStatusFrame_PulseWidth = 4
    kStatusFrame_MotionProfile = 5
    
    # Motion Profile status bits
    kMotionProfileFlag_ActTraj_IsValid = 0x1
    kMotionProfileFlag_HasUnderrun = 0x2
    kMotionProfileFlag_IsUnderrun = 0x4
    kMotionProfileFlag_ActTraj_IsLast = 0x8
    kMotionProfileFlag_ActTraj_VelOnly = 0x10
    
    # Motor output is neutral, Motion Profile Executer is not running.
    kMotionProfile_Disable = 0
    # Motor output is updated from Motion Profile Executer, MPE will
    # process the buffered points.
    kMotionProfile_Enable = 1
    # Motor output is updated from Motion Profile Executer, MPE will
    # stay processing current trajectory point.
    kMotionProfile_Hold = 2



# validated: 2018-01-05 EN 46d3d290a0b2 java/src/com/ctre/phoenix/motorcontrol/can/WPI_VictorSPX.java
#----------------------------------------------------------------------------
#  Software License Agreement
#
# Copyright (C) Cross The Road Electronics.  All rights
# reserved.
# 
# Cross The Road Electronics (CTRE) licenses to you the right to 
# use, publish, and distribute copies of CRF (Cross The Road) firmware files (*.crf) and Software
# API Libraries ONLY when in use with Cross The Road Electronics hardware products.
# 
# THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT
# LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL
# CROSS THE ROAD ELECTRONICS BE LIABLE FOR ANY INCIDENTAL, SPECIAL, 
# INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF
# PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
# BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE
# THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER
# SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT
# (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE
#----------------------------------------------------------------------------
import hal
from wpilib import MotorSafety, LiveWindow, SendableBase
from wpilib._impl.utils import match_arglist
from .victorspx import VictorSPX
from ._impl import ControlMode


__all__ = ['WPI_VictorSPX']


class WPI_VictorSPX (VictorSPX, SendableBase, MotorSafety):
    """WPI Compliant motor controller class.
    WPILIB's object model requires many interfaces to be implemented to use
    the various features.
    This includes...

    - Software PID loops running in the robot controller
    - LiveWindow/Test mode features
    - Motor Safety (auto-turn off of motor if Set stops getting called)
    - Single Parameter set that assumes a simple motor controller.
    """

    def __init__(self, deviceNumber: int):
        super().__init__(deviceNumber)
        hal.report(hal.UsageReporting.kResourceType_CTRE_future3, deviceNumber + 1)
        self.description = "Victor SPX %s" % (deviceNumber,)

        MotorSafety.__init__(self)
        self.setExpiration(0.0)
        self.setSafetyEnabled(False)

        SendableBase.__init__(self)
        LiveWindow.add(self)
        self.setName("Victor SPX ", deviceNumber)
        self.speed = 0.0

    def set(self, *args, **kwargs):
        """
        See :meth:`.BaseMotorController.set`

        Can be called three ways:

        - speed
        - mode, value
        - mode, demand0, demand1

        largely a wrapper around :meth:`.BaseMotorController.set`

        :param value:
        :type value: float
        :param mode: ControlMode.PercentOutput if not provided
        :type mode: ControlMode
        :param speed:
        :type speed: float
        :param demand0:
        :type demand0: float
        :param demand1:
        :type demand1: float
        """
        speed_arg = ("speed", [float, int])
        value_arg = ("value", [float, int])
        mode_arg = ("mode", [ControlMode])
        demand0_arg = ("demand0", [float, int])
        demand1_arg = ("demand1", [float, int])

        templates = [
            [speed_arg],
            [mode_arg, value_arg],
            [mode_arg, demand0_arg, demand1_arg]]

        index, results = match_arglist('WPI_VictorSPX.__init__',
                                   args, kwargs, templates)

        if index == 2:
            super().set(results['mode'], results['demand0'], results['demand1'])
        else:
            if index == 0:
                self.speed = value = results['speed']
                mode = ControlMode.PercentOutput
            elif index == 1:
                value = results['value']
                mode = results['mode']
            super().set(mode, value)

        self.feed()

    def pidWrite(self, output: float):
        self.set(output)

    def get(self) -> float:
        """Common interface for getting the current set speed of a speed controller.
        
        :returns: The current set speed. Value is between -1.0 and 1.0.
        """
        return self.speed

    #def setInverted(self, isInverted: bool):
    #    super().setInverted(isInverted)

    #def getInverted(self):
    #    return super().getInverted()

    def disable(self):
        self.neutralOutput()

    def stopMotor(self):
        """Common interface to stop the motor until :meth:`.set` is called again."""
        self.neutralOutput()

    def initSendable(self, builder):
        builder.setSmartDashboardType("Speed Controller")
        builder.setSafeState(self.stopMotor)
        builder.addDoubleProperty("Value", self.get, self.set)

    def getDescription(self):
        return self.description




#
# Much of this copied from https://github.com/pybind/python_example.git
#

import os
from os.path import dirname, exists, join
from setuptools import find_packages, setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist
import shutil
import subprocess
import sys
import setuptools

ctre_lib_version = '5.1.3.1'

setup_dir = dirname(__file__)
git_dir = join(setup_dir, '.git')
base_package = 'ctre'
version_file = join(setup_dir, base_package, 'version.py')

# Automatically generate a version.py based on the git version
if exists(git_dir):
    p = subprocess.Popen(["git", "describe", "--tags", "--long", "--dirty=-dirty"],
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()
    # Make sure the git version has at least one tag
    if err:
        print("Error: You need to create a tag for this repo to use the builder")
        sys.exit(1)

    # Convert git version to PEP440 compliant version
    # - Older versions of pip choke on local identifiers, so we can't include the git commit
    v, commits, local = out.decode('utf-8').rstrip().split('-', 2)
    if commits != '0' or '-dirty' in local:
        v = '%s.post0.dev%s' % (v, commits)

    # Create the version.py file
    with open(version_file, 'w') as fp:
        fp.write("# Autogenerated by setup.py\n__version__ = '{0}'".format(v))
        fp.write("\n__ctre_version__ = '%s'" % ctre_lib_version)

if exists(version_file):
    with open(join(setup_dir, base_package, 'version.py'), 'r') as fp:
        exec(fp.read(), globals())
else:
    __version__ = 'master'

with open(join(setup_dir, 'README.rst'), 'r') as readme_file:
    long_description = readme_file.read()


#
# pybind-specific compilation stuff
#

class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)

# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % ctre_lib_version)
            opts.append('-s') # strip
            opts.append('-g0') # remove debug symbols
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % ctre_lib_version)
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


install_requires = ['wpilib>=2018.0.0,<2019.0.0']

class Downloader:
    '''
        Utility object to allow lazily retrieving needed artifacts on demand,
        instead of distributing several extra MB with the pypi build.
    '''

    def __init__(self):
        self._halsrc = None
        self._ctresrc = None

        ctre_devdir = os.environ.get("RPY_CTRE_DEVDIR")
        if ctre_devdir:
            # development use only -- preextracted files so it doesn't have
            # to download it over and over again
            # -> if the directory doesn't exist, it will download the current
            #    files to that directory

            self._halsrc = join(ctre_devdir, 'hal')
            self._ctresrc = join(ctre_devdir, 'ctre')

    # copy/paste from hal_impl.distutils
    def _download(self, url):
        import atexit
        import posixpath
        from urllib.request import urlretrieve, urlcleanup
        import sys

        print("Downloading", posixpath.basename(url))

        def _reporthook(count, blocksize, totalsize):
            percent = int(count*blocksize*100/totalsize)
            sys.stdout.write("\r%02d%%" % percent)
            sys.stdout.flush()

        filename, _ = urlretrieve(url, reporthook=_reporthook)
        atexit.register(urlcleanup)
        return filename

    def _download_and_extract_zip(self, url, to=None):
        import atexit
        import shutil
        import tempfile
        import zipfile

        if to is None:
            # generate temporary directory
            tod = tempfile.TemporaryDirectory()
            to = tod.name
            atexit.register(tod.cleanup)

        zip_fname = self._download(url)
        with zipfile.ZipFile(zip_fname) as z:
            if isinstance(to, str):
                z.extractall(to)
                return to
            else:
                for src, dst in to.items():
                    with z.open(src, 'r') as zfp:
                        with open(dst, 'wb') as fp:
                            shutil.copyfileobj(zfp, fp)

    @property
    def halsrc(self):
        if not self._halsrc or not exists(self._halsrc):
            import hal_impl.distutils
            self._halsrc = hal_impl.distutils.extract_hal_libs(to=self._halsrc)
        return self._halsrc

    @property
    def ctresrc(self):
        if not self._ctresrc or not exists(self._ctresrc):
            url = 'http://www.ctr-electronics.com/downloads/lib/CTRE_Phoenix_FRCLibs_NON-WINDOWS_v%s.zip' % ctre_lib_version
            self._ctresrc = self._download_and_extract_zip(url, to=self._ctresrc)
        return self._ctresrc

get = Downloader()

# Detect roboRIO.. not foolproof, but good enough
if exists('/etc/natinst/share/scs_imagemetadata.ini'):


    ext_modules = [
        Extension(
            'ctre._impl.ctre_roborio',
            ['ctre/_impl/ctre_roborio.cpp'],
            include_dirs=[
                # Path to pybind11 headers
                get_pybind_include(),
                get_pybind_include(user=True),
                join(get.ctresrc, 'cpp', 'include'),
            ],
            libraries=['wpiHal', 'CTRE_PhoenixCCI'],
            library_dirs=[
                join(get.halsrc, 'linux', 'athena', 'shared'),
                join(get.ctresrc, 'cpp', 'lib'),
            ],
            language='c++',
        ),
    ]

    # This doesn't actually work, as it needs to be installed before setup.py is ran
    # ... but we specify it
    #install_requires = ['pybind11>=1.7']
    install_requires.append('robotpy-hal-roborio>=2018.0.0,<2019.0.0')
    cmdclass = {'build_ext': BuildExt}
else:
    install_requires.append('robotpy-hal-sim>=2018.0.0,<2019.0.0')
    ext_modules = None
    cmdclass = {}

#
# Autogenerating the required CTRE files is something that
# is done at sdist time. This means if you are testing builds,
# you have to run 'setup.py sdist build'.
#
# The advantage of doing it this way is that the autogen files
# are distributed with the pypi package, so simulation users
# don't have to install anything special to build this
#

class SDist(sdist):
    def run(self):
        from header2whatever import batch_convert

        # Do this before deleting the autogen directory, as it may fail
        ctresrc = get.ctresrc

        config_path = join(setup_dir, 'gen', 'gen.yml')
        outdir = join(setup_dir, 'ctre', '_impl', 'autogen')

        shutil.rmtree(outdir, ignore_errors=True)

        batch_convert(config_path, outdir, ctresrc)

        with open(join(outdir, '__init__.py'), 'w'):
            pass

        super().run()

cmdclass['sdist'] = SDist

setup(
    name='robotpy-ctre',
    version=__version__,
    author='Dustin Spicuzza',
    author_email='dustin@virtualroadside.com',
    url='https://github.com/robotpy/robotpy-ctre',
    description='RobotPy bindings for CTRE third party libraries',
    long_description=long_description,
    packages=find_packages(),
    ext_modules=ext_modules,
    install_requires=install_requires,
    cmdclass=cmdclass,
    zip_safe=False,
    entry_points={
        'robotpylib': ['info = ctre._impl.info:Info'],
        'robotpysim': ['ctre = ctre._impl.sim_ui:CtreUI']}
)
